"""The Countries module.

It exports classes related to countries CSV data:
    * Country
        DTO class representing a single row in countries CSV
        data.
    * CountriesDB
        Class exposing READ operations on countries CSV data.
"""
from __future__ import annotations

from typing import ClassVar

import pandas
from pandas import DataFrame

from mimeo import tools
from mimeo.database.exc import InvalidIndexError


class Country:
    """DTO class representing a single row in cities CSV data.

    Attributes
    ----------
    iso_3 : str
        A country ISO3 code
    iso_2 : str
        A country ISO2 code
    name : str
        A country name
    """

    def __init__(
            self,
            iso_3: str,
            iso_2: str,
            name: str,
    ):
        """Initialize Country class.

        Parameters
        ----------
        iso_3 : str
            A country ISO3 code
        iso_2 : str
            A country ISO2 code
        name : str
            A country name
        """
        self.iso_3: str = iso_3
        self.iso_2: str = iso_2
        self.name: str = name

    def __str__(
            self,
    ) -> str:
        """Stringify the Country instance.

        Returns
        -------
        str
            A stringified `dict` representation of the Country instance
        """
        return str({
            "iso_3": self.iso_3,
            "iso_2": self.iso_2,
            "name": self.name,
        })

    def __repr__(
            self,
    ) -> str:
        """Represent the Country instance.

        Returns
        -------
        str
            A python representation of the Country instance
        """
        return (f"Country("
                f"iso_3='{self.iso_3}', "
                f"iso_2='{self.iso_2}', "
                f"name='{self.name}')")


class CountriesDB:
    """Class exposing READ operations on countries CSV data.

    Attributes
    ----------
    NUM_OF_RECORDS : int
        A number of rows in countries CSV data

    Methods
    -------
    get_countries() -> list[Country]
        Get all countries.
    get_country_at(index: int) -> Country
        Get a country at `index` position.
    get_country_by_iso_3(iso_3: str) -> Country
        Get a country having a specific ISO3 code.
    get_country_by_iso_3(iso_2: str) -> Country
        Get a country having a specific ISO2 code.
    get_country_by_name(name: str) -> Country
        Get a country having a specific name.
    """

    NUM_OF_RECORDS: int = 239
    _COUNTRIES_DB: str = "countries.csv"
    _COUNTRIES_DF: DataFrame = None
    _COUNTRIES: ClassVar[list] = None

    def get_country_at(
            self,
            index: int,
    ) -> Country:
        """Get a country at `index` position.

        Parameters
        ----------
        index : int
            A country row index

        Returns
        -------
        Country
            A specific country

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        countries = CountriesDB._get_countries()
        try:
            return countries[index]
        except IndexError:
            last_index = CountriesDB.NUM_OF_RECORDS-1
            raise InvalidIndexError(index, last_index) from IndexError

    def get_country_by_iso_3(
            self,
            iso_3: str,
    ) -> Country:
        """Get a country having a specific ISO3 code.

        Parameters
        ----------
        iso_3 : str
            An ISO3 code to find a country

        Returns
        -------
        Country
            A specific country or None
        """
        countries = CountriesDB._get_countries()
        return next(filter(lambda country: country.iso_3 == iso_3, countries), None)

    def get_country_by_iso_2(
            self,
            iso_2: str,
    ) -> Country:
        """Get a country having a specific ISO2 code.

        Parameters
        ----------
        iso_2 : str
            An ISO2 code to find a country

        Returns
        -------
        Country
            A specific country or None
        """
        countries = CountriesDB._get_countries()
        return next(filter(lambda country: country.iso_2 == iso_2, countries), None)

    def get_country_by_name(
            self,
            name: str,
    ) -> Country:
        """Get a country having a specific name.

        Parameters
        ----------
        name : str
            A name to find a country

        Returns
        -------
        Country
            A specific country or None
        """
        countries = CountriesDB._get_countries()
        return next(filter(lambda country: country.name == name, countries), None)

    def get_countries(
            self,
    ) -> list[Country]:
        """Get all countries.

        Returns
        -------
        list[Country]
            List of all countries
        """
        return CountriesDB._get_countries().copy()

    @classmethod
    def _get_countries(
            cls,
    ) -> list[Country]:
        """Get all countries from cache.

        The countries list is initialized for the first time and cached
        in internal class attribute.

        Returns
        -------
        list[Country]
            List of all countries
        """
        if cls._COUNTRIES is None:
            cls._COUNTRIES = [Country(row.ISO_3, row.ISO_2, row.NAME)
                              for row in cls._get_countries_df().itertuples()]
        return cls._COUNTRIES

    @classmethod
    def _get_countries_df(
            cls,
    ) -> pandas.DataFrame:
        """Load countries CSV data and save in internal class attribute."""
        if cls._COUNTRIES_DF is None:
            data = tools.get_resource(cls._COUNTRIES_DB)
            cls._COUNTRIES_DF = pandas.read_csv(data)
        return cls._COUNTRIES_DF



"""The Mimeo Meta package.

This package is meant to store all useful classes that are not
essentially related to any Mimeo component.

It contains the following modules:
* alive
    The Alive module.
* exc
    The Mimeo Meta Exceptions module.

The Mimeo Meta package exports the following classes:
- OnlyOneAlive
    A type ensuring there's only one instance qualified to be used.
- Alive
    A superclass for OnlyOneAlive classes.

To use this package, simply import the desired class:
    from mimeo.meta import Alive, OnlyOneAlive
"""
from __future__ import annotations

from .alive import Alive, OnlyOneAlive

__all__ = ["Alive", "OnlyOneAlive"]



"""The Mimeo CLI Exceptions module.

It contains all custom exceptions related to Mimeo CLI:
    * PathNotFoundError
        A custom Exception class for not found path.
    * EnvironmentNotFoundError
        A custom Exception class for not found environment.
    * EnvironmentsFileNotFoundError
        A custom Exception class for not found environments file.
"""


from __future__ import annotations


class PathNotFoundError(Exception):
    """A custom Exception class for not found path.

    Raised while attempting to access a Mimeo Config file or directory that doesn't
    exist.
    """

    def __init__(
            self,
            path: str,
    ):
        """Initialize PathNotFoundError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        path : str
            A non existing path
        """
        super().__init__(f"No such file or directory [{path}]")


class EnvironmentNotFoundError(Exception):
    """A custom Exception class for not found environment.

    Raised while attempting to access an environment that does not exist.
    """

    def __init__(
            self,
            env_name: str,
            envs_file_path: str,
    ):
        """Initialize EnvironmentNotFoundError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        env_name : str
            An environment name
        envs_file_path : str
            An environments file path
        """
        super().__init__(f"No such env [{env_name}] "
                         f"in environments file [{envs_file_path}]")


class EnvironmentsFileNotFoundError(Exception):
    """A custom Exception class for not found environments file.

    Raised while attempting to access an environments file that does not exist.
    """

    def __init__(
            self,
            envs_file_path: str,
    ):
        """Initialize EnvironmentsFileNotFoundError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        envs_file_path : str
            An environments file path
        """
        super().__init__(f"Environments file not found [{envs_file_path}]")



"""The Mimeo Main module.

This module provides a CLI interface for the Mimeo.
It exports a single function:
    * main()
        Execute a Mimeo Job using CLI.
"""
from __future__ import annotations

from mimeo.cli import MimeoJob
from mimeo.logging import setup_logging


def main():
    """Execute a Mimeo Job using CLI."""
    setup_logging()
    MimeoJob().run()


if __name__ == "__main__":
    main()



"""The Mimeo Context Exceptions module.

It contains all custom exceptions related to Mimeo Context:
    * MinimumIdentifierReachedError
        A custom Exception class for reaching minimum identifier.
    * UninitializedContextIterationError
        A custom Exception class for uninitialized context's iteration.
    * ContextIterationNotFoundError
        A custom Exception class for not found context's iteration.
    * InvalidSpecialFieldNameError
        A custom Exception class for invalid special field's name.
    * InvalidSpecialFieldValueError
        A custom Exception class for invalid special field's value.
    * SpecialFieldNotFoundError
        A custom Exception class for not found special field.
    * VarNotFoundError
        A custom Exception class for not found var.
    * InvalidReferenceValueError
        A custom Exception class for invalid reference's value.
    * ReferenceNotFoundError
        A custom Exception class for not found reference.
    * NonPopulatedReferenceError
        A custom Exception class for non-populated reference.
    * NoCorrespondingReferenceError
        A custom Exception class for no corresponding reference.
"""
from __future__ import annotations


class MinimumIdentifierReachedError(Exception):
    """A custom Exception class for reaching minimum identifier.

    Raised when using MimeoContext.prev_id() method and id is equal to
    0.
    """

    def __init__(
            self,
    ):
        """Initialize MinimumIdentifierReachedError exception with details.

        Extends Exception constructor with a constant message.
        """
        super().__init__("There's no previous ID!")


class UninitializedContextIterationError(Exception):
    """A custom Exception class for uninitialized context's iteration.

    Raised while attempting to access the current iteration without
    prior initialization.
    """

    def __init__(
            self,
            context_name: str,
    ):
        """Initialize UninitializedContextIterationError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        context_name : str
            A current context name
        """
        msg = ("No iteration has been initialized "
               f"for the current context [{context_name}]")
        super().__init__(msg)


class ContextIterationNotFoundError(Exception):
    """A custom Exception class for not found context's iteration.

    Raised while attempting to access an iteration that does not exist.
    """

    def __init__(
            self,
            iteration_id: int,
            context_name: str,
    ):
        """Initialize ContextIterationNotFoundError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        iteration_id : int
            A current context name
        context_name : str
            A current context name
        """
        msg = (f"No iteration with id [{iteration_id}] "
               f"has been initialized for the current context [{context_name}]")
        super().__init__(msg)


class InvalidSpecialFieldNameError(Exception):
    """A custom Exception class for invalid special field's name.

    Raised while attempting to save a special field and provided name
    is not a string value.
    """

    def __init__(
            self,
    ):
        """Initialize InvalidSpecialFieldNameError exception with details.

        Extends Exception constructor with a constant message.
        """
        super().__init__("A special field name needs to be a string value!")


class InvalidSpecialFieldValueError(Exception):
    """A custom Exception class for invalid special field's value.

    Raised while attempting to save a special field and provided value
    is non-atomic one.
    """

    def __init__(
            self,
            field_value: dict | list,
    ):
        """Initialize InvalidSpecialFieldValueError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        field_value : dict | list
            A special field value
        """
        msg = f"Provided field value [{field_value}] is invalid (use any atomic value)!"
        super().__init__(msg)


class SpecialFieldNotFoundError(Exception):
    """A custom Exception class for not found special field.

    Raised while attempting to access a special field that does not
    exist.
    """

    def __init__(
            self,
            field_name: str,
    ):
        """Initialize SpecialFieldNotFoundError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        field_name : str
            A special field name
        """
        super().__init__(f"Special Field [{field_name}] has not been found!")


class VarNotFoundError(Exception):
    """A custom Exception class for not found var.

    Raised while attempting to access a variable that does not exist.
    """

    def __init__(
            self,
            variable_name: str,
    ):
        """Initialize VarNotFoundError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        variable_name : str
            A variable name
        """
        super().__init__(f"Provided variable [{variable_name}] is not defined!")


class InvalidReferenceValueError(Exception):
    """A custom Exception class for invalid reference's value.

    Raised while attempting to cache a reference value and provided value
    is non-atomic one.
    """

    def __init__(
            self,
            ref_name: dict | list,
    ):
        """Initialize InvalidReferenceValueError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        ref_name : dict | list
            A reference value
        """
        msg = (f"Provided reference value [{ref_name}] is invalid "
               "(use any atomic value)!")
        super().__init__(msg)


class ReferenceNotFoundError(Exception):
    """A custom Exception class for not found reference.

    Raised while attempting to access a reference that does not exist.
    """

    def __init__(
            self,
            ref_name: str,
    ):
        """Initialize ReferenceNotFoundError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        ref_name : str
            A reference name
        """
        super().__init__(f"Reference [{ref_name}] has not been found!")


class NonPopulatedReferenceError(Exception):
    """A custom Exception class for non-populated reference.

    Raised while attempting to access a reference that is not populated with any value.
    """

    def __init__(
            self,
            ref_name: str,
    ):
        """Initialize NonPopulatedReferenceError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        ref_name : str
            A reference name
        """
        super().__init__(f"Reference [{ref_name}] has not been populated "
                         "with any value!")


class NoCorrespondingReferenceError(Exception):
    """A custom Exception class for no corresponding reference.

    Raised while attempting to access a parallel reference that does not have
    a corresponding value at the same index.
    """

    def __init__(
            self,
            ref_name: str,
            iteration: int,
    ):
        """Initialize NoCorrespondingReferenceError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        ref_name : str
            A reference name
        iteration : int
            An iteration ID
        """
        super().__init__(f"No corresponding reference [{ref_name}] "
                         f"for the iteration [{iteration}].")



"""The Mimeo Iteration module.

It exports only one class:
    * MimeoIteration
        A class representing a single iteration in a Mimeo Template.
"""
from __future__ import annotations

import uuid

from mimeo.context.exc import (InvalidSpecialFieldNameError,
                               InvalidSpecialFieldValueError,
                               SpecialFieldNotFoundError)


class MimeoIteration:
    """A class representing a single iteration in a Mimeo Template.

    Each iteration has its own id (an ordinal number in a context),
    a key being a unique value across all iterations, and
    it stores special fields that could be ingested in other fields.

    Attributes
    ----------
    id : int
        An ordinal number in a Mimeo Context
    key : str
        An UUID value unique across all templates

    Methods
    -------
    add_special_field(self, field_name: str, field_value: str | int | bool)
        Put a special field entry to memory.
    get_special_field(self, field_name: str) -> str | int | bool]
        Get a special field value from memory.
    """

    def __init__(
            self,
            identifier: int,
    ):
        """Initialize MimeoIteration class.

        Parameters
        ----------
        identifier : int
            An ordinal number in a Mimeo Context
        """
        self.id: identifier = identifier
        self.key: str = str(uuid.uuid4())
        self._special_fields: dict = {}

    def add_special_field(
            self,
            field_name: str,
            field_value: str | int | bool,
    ):
        """Put a special field entry to memory.

        Parameters
        ----------
        field_name : str
            A special field name
        field_value : str | int | bool
            A special field value

        Raises
        ------
        InvalidSpecialFieldNameError
            If the special field name is not a string
        InvalidSpecialFieldValueError
            If the special field value is dict or list
        """
        if not isinstance(field_name, str):
            raise InvalidSpecialFieldNameError
        if isinstance(field_value, (dict, list)):
            raise InvalidSpecialFieldValueError(field_value)

        self._special_fields[field_name] = field_value

    def get_special_field(
            self,
            field_name: str,
    ) -> str | int | bool:
        """Get a special field value from memory.

        Parameters
        ----------
        field_name : str
            A special field name

        Returns
        -------
        str | int | bool
            A special field value

        Raises
        ------
        SpecialFieldNotFoundError
            If the special field does not exist.
        """
        if field_name not in self._special_fields:
            raise SpecialFieldNotFoundError(field_name)

        return self._special_fields.get(field_name)



"""The Mimeo Consumer module.

It exports only one class:
    * Consumer
        An abstract class for data consumers in Mimeo.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Collection, Generator


class Consumer(metaclass=ABCMeta):
    """An abstract class for data consumers in Mimeo.

    Its subclasses are meant to be used in the Mimeo processing.
    Every supported output direction has a Consumer representation.

    Methods
    -------
    consume
        Consumes data generated by Mimeo.
    """

    @classmethod
    def __subclasshook__(
            cls,
            subclass: Consumer,
    ):
        """Verify if a subclass implements all abstract methods.

        Parameters
        ----------
        subclass : Consumer
            A Consumer subclass

        Returns
        -------
        bool
            True if the subclass includes the consume method
        """
        return "consume" in subclass.__dict__ and callable(subclass.consume)

    @abstractmethod
    def consume(
            self,
            data: Collection | Generator,
    ):
        """Consume data generated by Mimeo.

        It is an abstract method to implement in subclasses

        Parameters
        ----------
        data : Collection | Generator
            Stringified data generated by Mimeo
        """
        raise NotImplementedError



"""The Mimeo Generator module.

It exports only one class:
    * Generator
        An abstract class for data generators in Mimeo.
"""
from __future__ import annotations

import logging
import xml.etree.ElementTree as ElemTree
from abc import ABCMeta, abstractmethod
from typing import Any, Iterator

from mimeo.config.mimeo_config import MimeoTemplate
from mimeo.context import MimeoContext, MimeoContextManager
from mimeo.context.decorators import (mimeo_clear_iterations, mimeo_context,
                                      mimeo_context_switch,
                                      mimeo_next_iteration)
from mimeo.utils import MimeoRenderer

logger = logging.getLogger(__name__)


class Generator(metaclass=ABCMeta):
    """An abstract class for data generators in Mimeo.

    Its subclasses are meant to be used in the Mimeo processing.
    Every supported output format has a Generator representation.

    Methods
    -------
    generate(
        templates: list | Iterator[MimeoTemplate],
        parent: Any = None
    ) -> Iterator[Any]
        Generate data based on the Mimeo Configuration.
    stringify(
        data_unit: Any
    ) -> str
        Stringify data generated by the generate() method.
    """

    @classmethod
    def __subclasshook__(
            cls,
            subclass: Generator,
    ):
        """Verify if a subclass implements all abstract methods.

        Parameters
        ----------
        subclass : Generator
            A Generator subclass

        Returns
        -------
        bool
            True if the subclass includes the generate and stringify
            methods
        """
        return ("generate" in subclass.__dict__ and callable(subclass.generate) and
                "stringify" in subclass.__dict__ and callable(subclass.stringify))

    @abstractmethod
    def generate(
            self,
            templates: list | Iterator[MimeoTemplate],
            parent: Any = None,
    ) -> Iterator[ElemTree.Element | dict]:
        """Generate data based on the Mimeo Configuration.

        It is an abstract method to implement in subclasses

        Parameters
        ----------
        templates : list | Iterator[MimeoTemplate]
            A collection of Mimeo Templates to process
        parent : Any, default None
            A parent node for the currently processed template.
            It is passed only when a Mimeo Config contain nested
            templates.

        Returns
        -------
        Iterator[ElemTree.Element | dict]
            Iterator for generated nodes
        """
        raise NotImplementedError

    @abstractmethod
    def stringify(
            self,
            data_unit: Any,
    ) -> str:
        """Stringify data generated by the generate() method.

        It is an abstract method to implement in subclasses

        Parameters
        ----------
        data_unit: Any
            A single data unit generated by the generate() method

        Returns
        -------
        str
            Stringified data unit
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _pre_process_node(
            cls,
            node_meta: dict,
    ) -> dict:
        """Pre-process node's metadata.

        This function adjusts existing node's metadata and completes it with custom
        properties.

        Parameters
        ----------
        node_meta : dict
            Initial node's metadata

        Returns
        -------
        dict
            Complete node's metadata
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _process_complex_value(
            cls,
            parent: ElemTree.Element | dict | list | None,
            node_meta: dict,
    ) -> ElemTree.Element | dict | list | None:
        """Process a node with a complex value.

        The node is processed accordingly to its value type.

        Parameters
        ----------
        parent : ElemTree.Element | dict | list | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        ElemTree.Element | dict | list | None
            A processed node
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _process_atomic_value(
            cls,
            parent: ElemTree.Element | dict | list | None,
            node_meta: dict,
    ) -> ElemTree.Element | dict | list | None:
        """Process a node with an atomic value.

        A parametrized Mimeo Util is considered as an atomic value as representing one.
        It renders a value for the node.

        Parameters
        ----------
        parent : ElemTree.Element | dict | list | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        ElemTree.Element | dict | list | None
            A processed node
        """
        raise NotImplementedError

    @classmethod
    @mimeo_context_switch
    @mimeo_clear_iterations
    def _process_single_template(
            cls,
            template: MimeoTemplate,
            parent: ElemTree.Element | dict | list | None = None,
    ) -> list[ElemTree.Element | dict]:
        """Process a single Mimeo Template.

        This function is used recursively when a Mimeo Configuration
        contains nested templates.
        It repeats same processing operation so many times as it is
        configured in the `count` property of a Mimeo Configuration.
        Before the template execution it switches context managed by
        MimeoContextManager and also clears iterations as nested
        templates would collect iterations from previous parent's
        iterations.

        Parameters
        ----------
        template : MimeoTemplate
            A single Mimeo Template to process
        parent : ElemTree.Element | dict | list | None, default None
            A parent node for processing nested templates

        Returns
        -------
        list[ElemTree.Element | dict]
            A list of generated data units
        """
        logger.debug("Reading template [%s]", template)
        return [cls._process_single_data_unit(template, parent)
                for _ in iter(range(template.count))]

    @classmethod
    @mimeo_next_iteration
    def _process_single_data_unit(
            cls,
            template: MimeoTemplate,
            parent: ElemTree.Element | dict | list | None = None,
    ) -> ElemTree.Element | dict:
        """Process a single data unit from the template.

        This function is used recursively when a Mimeo Configuration
        contains nested templates.
        It processes a single data unit. The reason why it is separated
        from the _process_node() function is the @mimeo_next_iteration
        decorator. The _process_node() function is recursively called
        by itself, and it produces data for a single iteration.
        The purpose of this function is to increment iteration before
        processing node.

        Parameters
        ----------
        template : MimeoTemplate
            A single Mimeo Template to process
        parent : ElemTree.Element | dict | list | None, default None
            A parent node for processing nested templates

        Returns
        -------
        ElemTree.Element | dict
            A single data unit generated within a single template
            iteration.
        """
        node_meta = cls._node_meta(
            template.model.root_name,
            template.model.root_data)
        return cls._process_node(parent, node_meta)

    @classmethod
    def _process_node(
            cls,
            parent: ElemTree.Element | dict | list | None,
            node_meta: dict,
    ) -> ElemTree.Element | dict:
        """Process a single template's node.

        This is a recursive function that traverses Mimeo Template and generates nodes
        based on node's metadata. First, element is pre-processed, in meaning
        of metadata being adjusted. Then, element is processed accordingly to its value
        type.

        Parameters
        ----------
        parent : ElemTree.Element | dict | list | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        ElemTree.Element | dict
            A single data unit generated within a single template iteration.
        """
        logger.fine("Rendering element - parent [%s], node_meta [%s]",
                    parent if not isinstance(parent, ElemTree.Element) else parent.tag,
                    node_meta)
        node_meta = cls._pre_process_node(node_meta)

        if cls._is_complex(node_meta):
            return cls._process_complex_value(parent, node_meta)
        return cls._process_atomic_value(parent, node_meta)

    @staticmethod
    def _is_complex(
            node_meta: dict,
    ) -> bool:
        """Verify if a node is complex.

        Parameters
        ----------
        node_meta : dict
            Node's metadata

        Returns
        -------
        bool
            True if node's value is a list or a dict not being a parametrized
            Mimeo Util. Otherwise, False.
        """
        return (isinstance(node_meta["value"], (list, dict)) and
                not node_meta["mimeo_util"])

    @staticmethod
    def _node_meta(
            name: str | None,
            value: dict | list | str | int | float | bool,
            attrs: dict | None = None,
            is_mimeo_util: bool | None = None,
            is_special_field: bool | None = None,
    ) -> dict:
        """Build node's metadata.

        Parameters
        ----------
        name : str | None
            A node's name (tag / property)
        value : dict | list | str | int | float | bool
            An node's value
        attrs : dict | None, default None
            An node's attributes
        is_mimeo_util : bool | None, default None
            A is-mimeo-util flag
        is_special_field : bool | None, default None
            A is-special-field flag

        Returns
        -------
        dict
            Entry's metadata
        """
        return {
            "name": name,
            "value": value,
            "attrs": attrs,
            "mimeo_util": is_mimeo_util,
            "special": is_special_field,
        }

    @staticmethod
    @mimeo_context
    def _render_atomic_value(
            node_meta: dict,
            context: MimeoContext | None = None,
    ) -> Any:
        """Render an atomic value for a node.

        It implements a common logic for all generators. Once the atomic value is
        rendered it is cached in references (if any configured) and special fields.

        Parameters
        ----------
        node_meta : dict
            Node's metadata
        context
            The current Mimeo Context (injected by MimeoContextManager)

        Returns
        -------
        value : Any
            A rendered value
        """
        value = MimeoRenderer.render(node_meta["value"])
        MimeoContextManager().cache_ref(node_meta["name"], value)
        if node_meta["special"]:
            context.curr_iteration().add_special_field(node_meta["name"], value)
        return value



"""The Mimeo Configuration Constants module."""
from __future__ import annotations

import yaml

from mimeo import tools

with tools.get_resource("constants.yaml") as config_file:
    constants = yaml.safe_load(config_file.read())
    _cc = constants["mimeo-config"]

CONFIG_XML_ROOT_NAME: str = _cc["key"]

########################################################################################
#                                    OUTPUT DETAILS                                    #
########################################################################################
_output_constants: dict = _cc["output"]
OUTPUT_KEY: str = _output_constants["key"]
OUTPUT_FORMAT_KEY: str = _output_constants["format"]["key"]
OUTPUT_DIRECTION_KEY: str = _output_constants["direction"]["key"]

########################################################################################
# --------------------------------- format specific ---------------------------------- #
_format_details: dict = _output_constants["format"]["details"]
OUTPUT_INDENT_KEY: str = _format_details["indent"]["key"]

_format_values: dict = _output_constants["format"]["values"]
SUPPORTED_OUTPUT_FORMATS: tuple = tuple(
    format_["key"]
    for format_ in _format_values.values())
OUTPUT_FORMAT_XML: str = _format_values["xml"]["key"]
OUTPUT_FORMAT_JSON: str = _format_values["json"]["key"]

# -------------------------------- xml format specific ------------------------------- #
_xml_format_details: dict = _format_values["xml"]["details"]
OUTPUT_XML_DECLARATION_KEY: str = _xml_format_details["xml-declaration"]["key"]

########################################################################################
# -------------------------------- direction specific -------------------------------- #
_direction_details: dict = _output_constants["direction"]["values"]
SUPPORTED_OUTPUT_DIRECTIONS: tuple = tuple(
    direction["key"]
    for direction in _direction_details.values())
OUTPUT_DIRECTION_FILE: str = _direction_details["file"]["key"]
OUTPUT_DIRECTION_STD_OUT: str = _direction_details["std-out"]["key"]
OUTPUT_DIRECTION_HTTP: str = _direction_details["http"]["key"]

# ----------------------------- file direction specific ------------------------------ #
_file_direction_details: dict = _direction_details["file"]["details"]
OUTPUT_DIRECTORY_PATH_KEY: str = _file_direction_details["directory-path"]["key"]
OUTPUT_FILE_NAME_KEY: str = _file_direction_details["file-name"]["key"]

# ----------------------------- http direction specific ------------------------------ #
_http_direction_details: dict = _direction_details["http"]["details"]
REQUIRED_HTTP_DETAILS: tuple = tuple(
    prop["key"] for prop in _http_direction_details.values()
    if prop.get("required", False) is True)

OUTPUT_METHOD_KEY: str = _http_direction_details["method"]["key"]
OUTPUT_PROTOCOL_KEY: str = _http_direction_details["protocol"]["key"]
OUTPUT_HOST_KEY: str = _http_direction_details["host"]["key"]
OUTPUT_PORT_KEY: str = _http_direction_details["port"]["key"]
OUTPUT_ENDPOINT_KEY: str = _http_direction_details["endpoint"]["key"]
OUTPUT_USERNAME_KEY: str = _http_direction_details["username"]["key"]
OUTPUT_PASSWORD_KEY: str = _http_direction_details["password"]["key"]

_req_method_details: dict = _http_direction_details["method"]["values"]
SUPPORTED_REQUEST_METHODS: tuple = tuple(_req_method_details.values())
OUTPUT_HTTP_REQUEST_POST: str = _req_method_details["post"]
OUTPUT_HTTP_REQUEST_PUT: str = _req_method_details["put"]

_req_protocol_details: dict = _http_direction_details["protocol"]["values"]
SUPPORTED_REQUEST_PROTOCOLS: tuple = tuple(_req_protocol_details.values())
OUTPUT_PROTOCOL_HTTP: str = _req_protocol_details["http"]
OUTPUT_PROTOCOL_HTTPS: str = _req_protocol_details["https"]

########################################################################################
#                                      MIMEO VARS                                      #
########################################################################################
_vars_constants: dict = _cc["vars"]
VARS_KEY: str = _vars_constants["key"]

########################################################################################
#                                      MIMEO REFS                                      #
########################################################################################
_refs_constants: dict = _cc["refs"]
REFS_KEY: str = _refs_constants["key"]

########################################################################################
# ---------------------------------- refs details ------------------------------------ #
_refs_details: dict = _refs_constants["details"]
REQUIRED_REFS_DETAILS: tuple = tuple(
    prop["key"] for prop in _refs_details.values()
    if prop.get("required", False) is True)
REFS_DETAIL_CONTEXT: str = _refs_details["context"]["key"]
REFS_DETAIL_FIELD: str = _refs_details["field"]["key"]
REFS_DETAIL_TYPE: str = _refs_details["type"]["key"]

# ------------------------------------ refs types ------------------------------------ #
_refs_types: dict = _refs_details["type"]["values"]
SUPPORTED_REFS_TYPES: tuple = tuple(_refs_types.values())
REFS_TYPE_ANY: str = _refs_details["type"]["values"]["any"]
REFS_TYPE_PARALLEL: str = _refs_details["type"]["values"]["parallel"]

########################################################################################
# ------------------------------ refs forbidden names -------------------------------- #
REFS_FORBIDDEN_NAMES: tuple = tuple(_refs_constants["forbidden-names"])

########################################################################################
#                                   MIMEO TEMPLATES                                    #
########################################################################################
_templates_constants: dict = _cc["templates"]
TEMPLATES_KEY: str = _templates_constants["key"]
TEMPLATES_XML_TEMPLATE_TAG: str = _templates_constants["xml-template-tag"]["key"]
TEMPLATES_COUNT_KEY: str = _templates_constants["count"]["key"]
TEMPLATES_MODEL_KEY: str = _templates_constants["model"]["key"]

########################################################################################
# -------------------------------- Mimeo Model level --------------------------------- #
_model_constants: dict = _templates_constants["model"]
MODEL_CONTEXT_KEY: str = _model_constants["context"]["key"]
MODEL_ATTRIBUTES_KEY: str = _model_constants["attributes"]["key"]
MODEL_TEXT_VALUE_KEY: str = _model_constants["text-node-value"]["key"]
MODEL_MIMEO_UTIL_KEY: str = _model_constants["mimeo-util"]["key"]
MODEL_MIMEO_UTIL_NAME_KEY: str = _model_constants["mimeo-util"]["name"]["key"]

__all__ = [
    "CONFIG_XML_ROOT_NAME",
    "OUTPUT_KEY",
    "OUTPUT_FORMAT_KEY",
    "OUTPUT_DIRECTION_KEY",
    "SUPPORTED_OUTPUT_FORMATS",
    "OUTPUT_FORMAT_XML",
    "OUTPUT_FORMAT_JSON",
    "OUTPUT_XML_DECLARATION_KEY",
    "OUTPUT_INDENT_KEY",
    "SUPPORTED_OUTPUT_DIRECTIONS",
    "OUTPUT_DIRECTION_FILE",
    "OUTPUT_DIRECTION_STD_OUT",
    "OUTPUT_DIRECTION_HTTP",
    "OUTPUT_DIRECTORY_PATH_KEY",
    "OUTPUT_FILE_NAME_KEY",
    "REQUIRED_HTTP_DETAILS",
    "OUTPUT_METHOD_KEY",
    "OUTPUT_PROTOCOL_KEY",
    "OUTPUT_HOST_KEY",
    "OUTPUT_PORT_KEY",
    "OUTPUT_ENDPOINT_KEY",
    "OUTPUT_USERNAME_KEY",
    "OUTPUT_PASSWORD_KEY",
    "SUPPORTED_REQUEST_METHODS",
    "OUTPUT_HTTP_REQUEST_POST",
    "OUTPUT_HTTP_REQUEST_PUT",
    "SUPPORTED_REQUEST_PROTOCOLS",
    "OUTPUT_PROTOCOL_HTTP",
    "OUTPUT_PROTOCOL_HTTPS",
    "VARS_KEY",
    "REFS_KEY",
    "REQUIRED_REFS_DETAILS",
    "REFS_DETAIL_CONTEXT",
    "REFS_DETAIL_FIELD",
    "REFS_DETAIL_TYPE",
    "SUPPORTED_REFS_TYPES",
    "REFS_TYPE_ANY",
    "REFS_TYPE_PARALLEL",
    "REFS_FORBIDDEN_NAMES",
    "TEMPLATES_KEY",
    "TEMPLATES_XML_TEMPLATE_TAG",
    "TEMPLATES_COUNT_KEY",
    "TEMPLATES_MODEL_KEY",
    "MODEL_CONTEXT_KEY",
    "MODEL_ATTRIBUTES_KEY",
    "MODEL_TEXT_VALUE_KEY",
    "MODEL_MIMEO_UTIL_KEY",
    "MODEL_MIMEO_UTIL_NAME_KEY",
]



"""The Mimeo package.

This package provides the Mimeo functionality with the highest level
class together with the Mimeo Configuration class:
* MimeoConfig
    A class representing Mimeo Configuration.
* MimeoConfigFactory
    A factory class to instantiate a MimeoConfig.
* Mimeograph
    A class responsible for the Mimeo processing.

It includes __main__.py module to provide a Command Line Interface
for the Mimeo.

This package is not meant to be used directly on low-level classes.
That's why we do not recommend importing particular packages, that are
internally used by Mimeograph:
* cli
    The Mimeo CLI package
* config
    The Mimeo Configuration package
* consumers
    The Mimeo Consumers package
* context
    The Mimeo Context package
* database
    The Mimeo Database package
* generators
    The Mimeo Generators package
* logging
    The Mimeo Logging package
* meta
    The Mimeo Meta package
* resources
    The Mimeo Resources package
* utils
    The Mimeo Utils package

The same apply to the following modules included by the Mimeo package:
* mimeo
    The Mimeo module
* tools
    The Mimeo Tools module
* exc
    The Mimeograph Exceptions module.

To use this package, simply import the desired classes:
    from mimeo import MimeoConfig, Mimeograph
"""
from __future__ import annotations

from .config import MimeoConfig, MimeoConfigFactory
from .mimeo import Mimeograph

__version__ = "1.1.0"
__all__ = ["MimeoConfig", "MimeoConfigFactory", "Mimeograph"]



"""The Mimeo Configuration module.

It contains classes representing Mimeo Configuration components at all levels.
All of them are Data Transfer Objects:
    * MimeoConfigFactory
        A factory class to instantiate a MimeoConfig.
    * MimeoDTO
        A superclass for all Mimeo configuration DTOs
    * MimeoConfig
        A MimeoDTO class representing Mimeo Configuration
    * MimeoOutput
        A MimeoDTO class representing Mimeo Output Details
    * MimeoTemplate
        A MimeoDTO class representing Mimeo Template
    * MimeoModel
        A MimeoDTO class representing Mimeo Model
"""
from __future__ import annotations

import json
import re
from pathlib import Path

import xmltodict

from mimeo.config import constants as cc
from mimeo.config.exc import (InvalidIndentError, InvalidMimeoConfigError,
                              InvalidMimeoModelError,
                              InvalidMimeoTemplateError, InvalidRefsError,
                              InvalidVarsError,
                              MimeoConfigurationNotFoundError,
                              MissingRequiredPropertyError,
                              UnsupportedMimeoConfigSourceError,
                              UnsupportedPropertyValueError)
from mimeo.logging import setup_logging

# setup logging when mimeo is used as a python library
setup_logging()


class MimeoConfigFactory:
    """A factory class to instantiate a MimeoConfig.

    Methods
    -------
    parse(source: dict | str) -> MimeoConfig
        Instantiate MimeoConfig from a source configuration.
    parse_source(source: str) -> dict
        Parse a Mimeo Configuration to a source dict.
    """

    @classmethod
    def parse(
            cls,
            source: dict | str,
    ) -> MimeoConfig:
        """Instantiate MimeoConfig from a source configuration.

        It uses internal methods to instantiate Mimeo Config depending on source type.

        Parameters
        ----------
        source : dict | str
            A source configuration

        Returns
        -------
        MimeoConfig
            A parsed MimeoConfig instance

        Raises
        ------
        UnsupportedMimeoConfigSourceError
            If the source is neither a dict nor str value
        """
        if isinstance(source, str):
            source = cls.parse_source(source)
        if not isinstance(source, dict):
            raise UnsupportedMimeoConfigSourceError(source)

        return MimeoConfig(source)

    @classmethod
    def parse_source(
            cls,
            source: str,
    ) -> dict:
        """Parse a Mimeo Configuration to a source dict.

        It uses internal methods to parse Mimeo Configuration depending on source value.

        Parameters
        ----------
        source : str
            A source configuration

        Returns
        -------
        dict
            A parsed source Mimeo Configuration ready to be used in MimeoConfig
            initialization.
        """
        if cls._is_file_path(source):
            source = cls._parse_source_from_file(source)
        else:
            source = cls._parse_source_from_str(source)
        return source

    @staticmethod
    def _is_file_path(
            source: str,
    ) -> bool:
        """Verify if the Mimeo Configuration source is a file path.

        Parameters
        ----------
        source : str
            A Mimeo Configuration source

        Returns
        -------
        bool
            True if the source is a file path. Otherwise, False.
        """
        return bool(re.match(r"([a-zA-Z0-9\s_/\\.\-\(\):])+(.json|.xml)$", source))

    @classmethod
    def _parse_source_from_file(
            cls,
            config_path: str,
    ) -> dict:
        """Parse a Mimeo Configuration file to a source dict.

        Parameters
        ----------
        config_path : str
            A source config file path

        Returns
        -------
        dict
            A parsed source Mimeo Configuration ready to be used in MimeoConfig
            initialization.
        """
        with Path(config_path).open() as config_file:
            if config_path.endswith(".json"):
                config = json.load(config_file)
            elif config_path.endswith(".xml"):
                config = cls._parse_source_from_str(config_file.read())
        return config

    @classmethod
    def _parse_source_from_str(
            cls,
            source: str,
    ) -> dict:
        """Parse a string Mimeo Configuration to a source dict.

        Parameters
        ----------
        source : str
            A source string to parse

        Returns
        -------
        dict
            A parsed source Mimeo Configuration ready to be used in MimeoConfig
            initialization.
        """
        parsed_source = xmltodict.parse(source)
        if cc.CONFIG_XML_ROOT_NAME not in parsed_source:
            source_key = list(parsed_source.keys())[0]
            raise MimeoConfigurationNotFoundError(source_key)
        parsed_source = parsed_source[cc.CONFIG_XML_ROOT_NAME]
        return cls._parse_source_values(parsed_source)

    @classmethod
    def _parse_source_values(
            cls,
            source_node: None | str | bool | int | float | dict | list,
    ) -> None | str | bool | int | float | dict | list:
        """Parse source values.

        This method recursively parses result of parsing XML to dict.

        Parameters
        ----------
        source_node : None | str | bool | int | float | dict | list
            An XML Mimeo Configuration's node parsed to dict

        Returns
        -------
        None | str | bool | int | float | dict | list
            An XML Mimeo Configuration's node with parsed nested values.
        """
        if isinstance(source_node, str):
            return cls._parse_str_source_value(source_node)
        if isinstance(source_node, list):
            return cls._parse_list_source_value(source_node)
        if isinstance(source_node, dict):
            return cls._parse_dict_source_value(source_node)
        return None

    @classmethod
    def _parse_str_source_value(
            cls,
            source_node: str,
    ) -> bool | float | int | str:
        """Parse a string value.

        This method returns boolean values for 'true' and 'false' strings and casts
        to float or int numeric ones. Otherwise, returns a source value.

        Parameters
        ----------
        source_node : str
            An XML Mimeo Configuration's node

        Returns
        -------
        bool | float | int | str
            A parsed string value

        Examples
        --------
        MimeoDTO._parse_str_source_value('true')
        -> True

        MimeoDTO._parse_str_source_value('false')
        -> False

        MimeoDTO._parse_str_source_value('1.5')
        -> 1.5

        MimeoDTO._parse_str_source_value('1')
        -> 1

        MimeoDTO._parse_str_source_value('value')
        -> 'value'
        """
        if source_node == "true":
            return True
        if source_node == "false":
            return False
        if source_node.replace("-", "").isnumeric():
            return int(source_node)
        if re.sub(r"[-\.]", "", source_node).isnumeric():
            return float(source_node)
        return source_node

    @classmethod
    def _parse_list_source_value(
            cls,
            source_node: list,
    ) -> list:
        """Parse a list value.

        This method parses all list items.

        Parameters
        ----------
        source_node : list
            An XML Mimeo Configuration's node

        Returns
        -------
        list
            A list with parsed items

        Examples
        --------
        MimeoDTO._parse_list_source_value(['true', 'false', '1.5', '1', 'value'])
        -> [True, False, 1.5, 1, 'value']
        """
        return [cls._parse_source_values(value) for value in source_node]

    @classmethod
    def _parse_dict_source_value(
            cls,
            source_node: dict,
    ) -> dict:
        """Parse a dict value.

        This method parses all dict values. Additionally, it applies a specific logic
        for Mimeo Templates and random_item Mimeo Util. As XML Mimeo Configuration needs
        a template node in templates, it assigns them to the "_templates_" key (moves
        one level up). Similar modification is made for items of the random_item Mimeo
        Util.

        Parameters
        ----------
        source_node : dict
            An XML Mimeo Configuration's node

        Returns
        -------
        dict
            A dict with parsed values

        Examples
        --------
        MimeoDTO._parse_dict_source_value({
            'SomeField1': 'true',
            'SomeField2': 'false',
            'SomeField3': '1.5',
            'SomeField4': '1',
            'SomeField5': 'value',
        })
        -> {
            'SomeField1': True,
            'SomeField2': False,
            'SomeField3': 1.5,
            'SomeField4': 1,
            'SomeField5': 'value',
        }
        """
        for key, value in source_node.items():
            if key == cc.TEMPLATES_KEY:
                cls._flatten_list(
                    source_node,
                    cc.TEMPLATES_KEY,
                    cc.TEMPLATES_XML_TEMPLATE_TAG)
            elif (key == cc.MODEL_MIMEO_UTIL_KEY and
                  value is not None and
                  value.get(cc.MODEL_MIMEO_UTIL_NAME_KEY) == "random_item"):
                cls._flatten_list(
                    source_node[key],
                    "items",
                    "item")
            else:
                source_node[key] = cls._parse_source_values(value)

        return source_node

    @classmethod
    def _flatten_list(
            cls,
            source_node: str | dict | list,
            key: str,
            child_key: str,
    ):
        """Move child node value one level up to a list.

        Parameters
        ----------
        source_node : str | dict | list
            An XML Mimeo Configuration's node
        key : str
            A parent key to being reassigned
        child_key : str
            A child key to take value from

        Examples
        --------
        MimeoDTO._parse_dict_source_value({
            '_templates_': None,
        })
        -> {
            '_templates_': [],
        }

        MimeoDTO._parse_dict_source_value({
            '_templates_': {
                '_template_' : {
                    'SomeField': 'true',
                },
            }
        })
        -> {
            '_templates_': [
                {
                    'SomeField': 'true',
                },
            ],
        }

        MimeoDTO._parse_dict_source_value({
            '_templates_': {
                '_template_' : [
                    {
                        'SomeField': 'true',
                    },
                    {
                        'SomeField': 'false',
                    },
                ],
            },
        })
        -> {
            '_templates_': [
               {
                   'SomeField': 'true',
               },
               {
                   'SomeField': 'false',
               },
            ],
        }

        MimeoDTO._parse_dict_source_value({
            '_name': 'random_item',
            'items': {
                'item': [
                    'value',
                    1,
                    True
                ]
            }
        })
        -> {
            '_name': 'random_item',
            'items': [
                'value',
                1,
                True
            ]
        }
        """
        value = source_node[key]
        if value is None:
            source_node[key] = []
        elif isinstance(value, list):
            source_node[key] = cls._parse_source_values(value)
        else:
            templates = source_node[key].get(child_key)
            if isinstance(templates, (str, dict)):
                source_node[key] = [cls._parse_source_values(templates)]
            elif isinstance(templates, list):
                source_node[key] = cls._parse_source_values(templates)
            else:
                source_node[key] = cls._parse_source_values(value)


class MimeoDTO:
    """A superclass for all Mimeo configuration DTOs.

    It is meant to store a source dictionary for logging purposes.

    Methods
    -------
    __str__() -> str
        Return the stringified source dictionary of a DTO.
    """

    def __init__(
            self, source: dict,
    ):
        """Initialize MimeoDTO class.

        Parameters
        ----------
        source : dict
            The source dictionary for a Mimeo DTO
        """
        self._source: dict = source

    def __str__(
            self,
    ) -> str:
        """Return the stringified source dictionary of a DTO."""
        return str(self._source)


class MimeoConfig(MimeoDTO):
    """A MimeoDTO class representing Mimeo Configuration.

    It is a python representation of a Mimeo Configuration file / dictionary.

    output : MimeoOutput, default {}
        A Mimeo Output Details settings
    vars : dict, default {}
        A Mimeo Configuration vars setting
    refs : dict, default {}
        A Mimeo Configuration refs setting
    templates : list
        A Mimeo Templates setting
    """

    def __init__(
            self,
            config: dict,
    ):
        """Initialize MimeoConfig class.

        Extends MimeoDTO constructor.

        Parameters
        ----------
        config : dict
            A source config dictionary
        """
        super().__init__(config)
        self.output: MimeoOutput = MimeoOutput(config.get(cc.OUTPUT_KEY, {}))
        self.vars: dict = self._get_vars(config)
        self.refs: dict = self._get_refs(config)
        self.templates: list[MimeoTemplate] = self._get_templates(config)

    @classmethod
    def _get_vars(
            cls,
            config: dict,
    ) -> dict:
        """Extract variables from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary

        Returns
        -------
        variables : dict
            Customized variables or an empty dictionary

        Raises
        ------
        InvalidVarsError
            If (1) the vars key does not point to a dictionary or
            (2) some variable's name does not start with a letter,
            is not SNAKE_UPPER_CASE with possible digits or
            (3) some variable's value points to non-atomic value nor Mimeo Util
        """
        variables = config.get(cc.VARS_KEY, {})
        if not isinstance(variables, dict):
            raise InvalidVarsError(InvalidVarsError.Code.ERR_1, vars=variables)
        for var, val in variables.items():
            if isinstance(val, (list, dict)) and not cls._is_mimeo_util_object(val):
                raise InvalidVarsError(InvalidVarsError.Code.ERR_2, var=var)
            if not re.match(r"^[A-Z][A-Z_0-9]*$", var):
                raise InvalidVarsError(InvalidVarsError.Code.ERR_3, var=var)
        return variables

    @classmethod
    def _get_refs(
            cls,
            config: dict,
    ) -> dict:
        """Extract references from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary

        Returns
        -------
        references : dict
            Customized references or an empty dictionary

        Raises
        ------
        InvalidRefsError
            If (1) the refs key does not point to a dictionary or
            (2) any ref is not a dictionary or
            (3) some refs does not have required details (context, field, type)
            (4) some refs are configured using names of Mimeo Utils or Vars
        UnsupportedPropertyValueError
            If the configured reference type is not supported
        """
        references = config.get(cc.REFS_KEY, {})
        if not isinstance(references, dict):
            raise InvalidRefsError(InvalidRefsError.Code.ERR_1, refs=references)

        variables = config.get(cc.VARS_KEY, {}).keys()
        missing_details_references = []
        forbidden_names_references = []
        for name, reference in references.items():
            if not isinstance(reference, dict):
                raise InvalidRefsError(InvalidRefsError.Code.ERR_2, ref=name)

            if any(detail not in reference for detail in cc.REQUIRED_REFS_DETAILS):
                missing_details_references.append(name)
            elif name in cc.REFS_FORBIDDEN_NAMES or name in variables:
                forbidden_names_references.append(name)
            else:
                ref_type = reference[cc.REFS_DETAIL_TYPE]
                if ref_type not in cc.SUPPORTED_REFS_TYPES:
                    raise UnsupportedPropertyValueError(
                        cc.REFS_DETAIL_TYPE,
                        ref_type,
                        cc.SUPPORTED_REFS_TYPES)
        if len(missing_details_references) > 0:
            raise InvalidRefsError(
                InvalidRefsError.Code.ERR_3,
                required=cc.REQUIRED_REFS_DETAILS,
                refs=missing_details_references)
        if len(forbidden_names_references) > 0:
            raise InvalidRefsError(
                InvalidRefsError.Code.ERR_4,
                refs=forbidden_names_references)

        return references

    @classmethod
    def _get_templates(
            cls,
            config: dict,
    ) -> list[MimeoTemplate]:
        """Extract Mimeo Templates from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary

        Returns
        -------
        list[MimeoTemplate]
            A Mimeo Templates list

        Raises
        ------
        InvalidMimeoConfigError
            If (1) the source dictionary does not include the _templates_ key or
            (2) the _templates_ key does not point to a list
        """
        templates = config.get(cc.TEMPLATES_KEY)
        if templates is None:
            raise InvalidMimeoConfigError(InvalidMimeoConfigError.Code.ERR_1,
                                          config=config)
        if not isinstance(templates, list):
            raise InvalidMimeoConfigError(InvalidMimeoConfigError.Code.ERR_2,
                                          config=config)
        return [MimeoTemplate(template)
                for template in config.get(cc.TEMPLATES_KEY)]

    @classmethod
    def _is_mimeo_util_object(
            cls,
            obj: dict,
    ) -> bool:
        """Verify if the object is a Mimeo Util.

        Parameters
        ----------
        obj : dict
            An object to verify

        Returns
        -------
        bool
            True if the object is a dictionary having only one key: _mimeo_util.
            Otherwise, False.
        """
        return (isinstance(obj, dict) and
                len(obj) == 1 and
                cc.MODEL_MIMEO_UTIL_KEY in obj)


class MimeoOutput(MimeoDTO):
    """A MimeoDTO class representing Mimeo Output Details.

    It is a python representation of a Mimeo Output Details configuration node.

    Attributes
    ----------
    direction : str, default 'file'
        The configured output direction
    format : str, default 'xml'
        A Mimeo Configuration output format setting
    xml_declaration : bool, default None
        A Mimeo Configuration xml declaration setting
    indent : int, default 0
        A Mimeo Configuration indent setting
    directory_path : str, default 'mimeo-output'
        The configured file output directory
    file_name : str, default 'mimeo-output-{}.{output_format}'
        The configured file output file name template
    method : str, default POST
        The configured http output request method
    protocol : str, default 'http'
        The configured http output protocol
    host : str
        The configured http output host
    port : str
        The configured http output port
    endpoint : str
        The configured http output endpoint
    username : str
        The configured http output username
    password : str
        The configured http output password
    """

    def __init__(
            self,
            output: dict,
    ):
        """Initialize MimeoOutput class.

        Extends MimeoDTO constructor.

        Parameters
        ----------
        output : dict
            A source config output details dictionary
        """
        super().__init__(output)
        self.direction: str = self._get_direction(output)
        self._validate_output(self.direction, output)
        self.format: str = self._get_format(output)
        self.xml_declaration: bool = self._get_xml_declaration(output, self.format)
        self.indent: int = self._get_indent(output)
        self.directory_path: str = self._get_directory_path(self.direction, output)
        self.file_name: str = self._get_file_name(self.direction, output, self.format)
        self.method: str = self._get_method(self.direction, output)
        self.protocol: str = self._get_protocol(self.direction, output)
        self.host: str = self._get_host(self.direction, output)
        self.port: int = self._get_port(self.direction, output)
        self.endpoint: str = self._get_endpoint(self.direction, output)
        self.username: str = self._get_username(self.direction, output)
        self.password: str = self._get_password(self.direction, output)

    @staticmethod
    def _get_direction(
            output: dict,
    ) -> str:
        """Extract output direction from the source dictionary.

        Parameters
        ----------
        output : dict
            A source config output details dictionary

        Returns
        -------
        direction : str
            The configured output direction

        Raises
        ------
        UnsupportedPropertyValueError
            If the configured output direction is not supported
        """
        direction = output.get(cc.OUTPUT_DIRECTION_KEY, cc.OUTPUT_DIRECTION_FILE)
        if direction not in cc.SUPPORTED_OUTPUT_DIRECTIONS:
            raise UnsupportedPropertyValueError(
                cc.OUTPUT_DIRECTION_KEY,
                direction,
                cc.SUPPORTED_OUTPUT_DIRECTIONS)
        return direction

    @staticmethod
    def _get_format(
            config: dict,
    ) -> str:
        """Extract an output format from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary

        Returns
        -------
        output_format : str
            The customized output format or 'xml' by default

        Raises
        ------
        UnsupportedPropertyValueError
            If the customized output format is not supported
        """
        output_format = config.get(cc.OUTPUT_FORMAT_KEY, cc.OUTPUT_FORMAT_XML)
        if output_format not in cc.SUPPORTED_OUTPUT_FORMATS:
            raise UnsupportedPropertyValueError(
                cc.OUTPUT_FORMAT_KEY,
                output_format,
                cc.SUPPORTED_OUTPUT_FORMATS)
        return output_format

    @staticmethod
    def _get_xml_declaration(
            config: dict,
            output_format: str,
    ) -> bool | None:
        """Extract an XML declaration setting from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary
        output_format : str
            The configured output format

        Returns
        -------
        bool | None
            The configured XML declaration setting when the output format is 'xml'.
            Otherwise, None. If the 'xml_declaration' setting is missing returns
            False by default.
        """
        if output_format == cc.OUTPUT_FORMAT_XML:
            return config.get(cc.OUTPUT_XML_DECLARATION_KEY, False)
        return None

    @staticmethod
    def _get_indent(
            config: dict,
    ) -> int:
        """Extract an indent value from the source dictionary.

        Parameters
        ----------
        config : dict
            A source config dictionary

        Returns
        -------
        indent : int
            The customized indent or 0 by default

        Raises
        ------
        InvalidIndentError
            If the customized indent is lower than zero
        """
        indent = config.get(cc.OUTPUT_INDENT_KEY, 0)
        if indent < 0:
            raise InvalidIndentError(indent)
        return indent

    @staticmethod
    def _get_directory_path(
            direction: str,
            output: dict,
    ) -> str | None:
        """Extract an output directory path from the source dictionary.

        It is extracted only when the output direction is 'file'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary

        Returns
        -------
        str | None
            The configured output directory path when the output direction is 'file'.
            Otherwise, None. If the 'directory_path' setting is missing returns
            'mimeo-output' by default.
        """
        if direction == cc.OUTPUT_DIRECTION_FILE:
            return output.get(cc.OUTPUT_DIRECTORY_PATH_KEY, "mimeo-output")
        return None

    @staticmethod
    def _get_file_name(
            direction: str,
            output: dict,
            output_format: str,
    ) -> str | None:
        """Generate an output file name template based on the source dictionary.

        It is generated only when the output direction is 'file'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary
        output_format : str
            The configured output format

        Returns
        -------
        str | None
            The configured output file name template when the output direction is
            'file'. Otherwise, None. If the 'file_name' setting is missing returns
            'mimeo-output-{}.{output_format}' by default.
        """
        if direction == cc.OUTPUT_DIRECTION_FILE:
            file_name = output.get(cc.OUTPUT_FILE_NAME_KEY, "mimeo-output")
            return f"{file_name}-{'{}'}.{output_format}"
        return None

    @staticmethod
    def _get_method(
            direction: str,
            output: dict,
    ) -> str | None:
        """Extract an HTTP request method from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary

        Returns
        -------
        method: str | None
            The configured HTTP request method when the output direction is 'http'.
            Otherwise, None. If the 'method' setting is missing returns
            'POST' by default.

        Raises
        ------
        UnsupportedPropertyValueError
            If the configured request method is not supported
        """
        method = None
        if direction == cc.OUTPUT_DIRECTION_HTTP:
            method = output.get(cc.OUTPUT_METHOD_KEY, cc.OUTPUT_HTTP_REQUEST_POST)
            if method not in cc.SUPPORTED_REQUEST_METHODS:
                raise UnsupportedPropertyValueError(
                    cc.OUTPUT_METHOD_KEY,
                    method,
                    cc.SUPPORTED_REQUEST_METHODS)
        return method

    @staticmethod
    def _get_protocol(
            direction: str,
            output: dict,
    ) -> str | None:
        """Extract an HTTP protocol from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary

        Returns
        -------
        str | None
            The configured HTTP request method when the output direction is 'http'.
            Otherwise, None. If the 'protocol' setting is missing returns
            'http' by default.

        Raises
        ------
        UnsupportedPropertyValueError
            If the configured request protocol is not supported
        """
        protocol = None
        if direction == cc.OUTPUT_DIRECTION_HTTP:
            protocol = output.get(cc.OUTPUT_PROTOCOL_KEY, cc.OUTPUT_PROTOCOL_HTTP)
            if protocol not in cc.SUPPORTED_REQUEST_PROTOCOLS:
                raise UnsupportedPropertyValueError(
                    cc.OUTPUT_PROTOCOL_KEY,
                    protocol,
                    cc.SUPPORTED_REQUEST_PROTOCOLS)
        return protocol

    @staticmethod
    def _get_host(
            direction: str,
            output: dict,
    ) -> str | None:
        """Extract an HTTP host from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary

        Returns
        -------
        str | None
            The configured HTTP host when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == cc.OUTPUT_DIRECTION_HTTP:
            return output.get(cc.OUTPUT_HOST_KEY)
        return None

    @staticmethod
    def _get_port(
            direction: str,
            output: dict,
    ) -> int | None:
        """Extract an HTTP port from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary

        Returns
        -------
        int | None
            The configured HTTP port when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == cc.OUTPUT_DIRECTION_HTTP:
            return output.get(cc.OUTPUT_PORT_KEY)
        return None

    @staticmethod
    def _get_endpoint(
            direction: str,
            output: dict,
    ) -> str | None:
        """Extract an HTTP endpoint from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary

        Returns
        -------
        str | None
            The configured HTTP request method when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == cc.OUTPUT_DIRECTION_HTTP:
            return output.get(cc.OUTPUT_ENDPOINT_KEY)
        return None

    @staticmethod
    def _get_username(
            direction: str,
            output: dict,
    ) -> str | None:
        """Extract a username from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary

        Returns
        -------
        str | None
            The configured username when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == cc.OUTPUT_DIRECTION_HTTP:
            return output.get(cc.OUTPUT_USERNAME_KEY)
        return None

    @staticmethod
    def _get_password(
            direction: str,
            output: dict,
    ) -> str | None:
        """Extract a password from the source dictionary.

        It is extracted only when the output direction is 'http'.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary

        Returns
        -------
        str | None
            The configured password when the output direction is 'http'.
            Otherwise, None.
        """
        if direction == cc.OUTPUT_DIRECTION_HTTP:
            return output.get(cc.OUTPUT_PASSWORD_KEY)
        return None

    @staticmethod
    def _validate_output(
            direction: str,
            output: dict,
    ) -> None:
        """Validate output details in the source dictionary.

        The validation is being done according to the configured output
        direction.

        Parameters
        ----------
        direction : str
            The configured output direction
        output : dict
            A source config output details dictionary

        Raises
        ------
        MissingRequiredPropertyError
            If the output details doesn't include all required settings
            for the direction
        """
        if direction == cc.OUTPUT_DIRECTION_HTTP:
            missing_details = [detail
                               for detail in cc.REQUIRED_HTTP_DETAILS
                               if detail not in output]
            if len(missing_details) > 0:
                raise MissingRequiredPropertyError(missing_details)


class MimeoTemplate(MimeoDTO):
    """A MimeoDTO class representing Mimeo Template.

    It is a python representation of a Mimeo Template configuration node.

    Attributes
    ----------
    count : int
        A configured count of the Mimeo Template
    model : MimeoModel
        A configured model of the Mimeo Template
    """

    def __init__(
            self,
            template: dict,
    ):
        """Initialize MimeoTemplate class.

        Extends MimeoDTO constructor.

        Parameters
        ----------
        template : dict
            A source config template dictionary
        """
        super().__init__(template)
        self._validate_template(template)
        self.count: int = template.get(cc.TEMPLATES_COUNT_KEY)
        self.model: MimeoModel = MimeoModel(template.get(cc.TEMPLATES_MODEL_KEY))

    @staticmethod
    def _validate_template(
            template: dict,
    ) -> None:
        """Validate template in the source dictionary.

        Parameters
        ----------
        template : dict
            A source config template dictionary

        Raises
        ------
        InvalidMimeoTemplateError
            If the source config doesn't include count or model properties
        """
        if cc.TEMPLATES_COUNT_KEY not in template:
            prop_name = "count"
            raise InvalidMimeoTemplateError(prop_name, template)
        if cc.TEMPLATES_MODEL_KEY not in template:
            prop_name = "model"
            raise InvalidMimeoTemplateError(prop_name, template)


class MimeoModel(MimeoDTO):
    """A MimeoDTO class representing Mimeo Model.

    It is a python representation of a Mimeo Model configuration node.

    Attributes
    ----------
    root_name : str
        A root node's tag
    root_data : dict
        A template data
    context_name : str
        A context name (root_name by default)
    """

    def __init__(
            self,
            model: dict,
    ):
        """Initialize MimeoModel class.

        Extends MimeoDTO constructor.

        Parameters
        ----------
        model : dict
            A source config model dictionary
        """
        super().__init__(model)
        self.root_name: str = MimeoModel._get_root_name(model)
        self.root_data: dict = model.get(self.root_name)
        self.context_name: str = MimeoModel._get_context_name(model, self.root_name)

    @staticmethod
    def _get_root_name(
            model: dict,
    ) -> str:
        """Extract a root name from the source dictionary.

        Parameters
        ----------
        model : dict
            A source config model dictionary

        Returns
        -------
        str
            The configured root node's tag

        Raises
        ------
        InvalidMimeoModelError
            If the source config has no or more than one root nodes
        """
        model_keys = list(filter(MimeoModel._is_not_configuration_key, iter(model)))
        if len(model_keys) == 0:
            raise InvalidMimeoModelError(InvalidMimeoModelError.Code.ERR_1, model=model)
        if len(model_keys) > 1:
            raise InvalidMimeoModelError(InvalidMimeoModelError.Code.ERR_2, model=model)
        return model_keys[0]

    @staticmethod
    def _get_context_name(
            model: dict,
            root_name: str,
    ) -> str:
        """Extract a context name from the source dictionary.

        Parameters
        ----------
        model : dict
            A source config model dictionary
        root_name : str
            The configured root node's tag

        Returns
        -------
        str
            The configured context name.
            If the 'context' setting is missing returns root name by default

        Raises
        ------
        InvalidMimeoModelError
            If the source config has a context name not being a string value
        """
        context_name = model.get(cc.MODEL_CONTEXT_KEY, root_name)
        if not isinstance(context_name, str):
            raise InvalidMimeoModelError(InvalidMimeoModelError.Code.ERR_3, model=model)
        return context_name

    @staticmethod
    def _is_not_configuration_key(
            dict_key: str,
    ) -> bool:
        """Verify if the dictionary key is a configuration one.

        Parameters
        ----------
        dict_key : str
            A dictionary key to verify

        Returns
        -------
        bool
            True if the key is 'context', otherwise False
        """
        return dict_key not in [cc.MODEL_CONTEXT_KEY]



"""The Mimeo HTTP Consumer module.

It exports only one class:
    * HttpConsumer
        A Consumer implementation sending data in HTTP requests.
"""
from __future__ import annotations

import asyncio
import logging
import uuid
from typing import Collection, Generator

from aiohttp import BasicAuth, ClientSession

from mimeo.config.mimeo_config import MimeoOutput
from mimeo.consumers import Consumer

logger = logging.getLogger(__name__)


class HttpConsumer(Consumer):
    """A Consumer implementation sending data in HTTP requests.

    This Consumer is instantiated for the 'http' output direction
    and sends data produced by Mimeo in an HTTP request body
    using Mimeo Output Details.

    Methods
    -------
    consume
        Send data generated by Mimeo in an HTTP request.

    Attributes
    ----------
    method : str
        An HTTP request method
    url : str
        An URL address to send the HTTP request
    """

    def __init__(
            self,
            output: MimeoOutput,
    ):
        """Initialize HTTPConsumer class.

        Parameters
        ----------
        output : MimeoOutput
            Configured Mimeo Output Details
        """
        self.method: str = output.method
        self.url: str = HttpConsumer.__build_url(output)
        self.__auth: BasicAuth = BasicAuth(output.username, output.password, "utf-8")
        self.__content_type: str = f"application/{output.format}"

    async def consume(
            self,
            data: Collection | Generator,
    ):
        """Send data generated by Mimeo in an HTTP request.

        It is an implementation of Consumer's abstract method.

        Parameters
        ----------
        data : Collection | Generator
            Stringified data generated by Mimeo
        """
        async def send_request(sess: ClientSession, data_unit: str):
            req_id = str(uuid.uuid4())
            logger.info("Sending request %s %s [%s]", self.method, self.url, req_id)
            resp = await sess.request(
                self.method,
                self.url,
                auth=self.__auth,
                data=data_unit,
                headers={"Content-Type": self.__content_type})
            logger.info("[%s] Status: %s", req_id, resp.status)

        async with ClientSession() as s:
            await asyncio.gather(*[send_request(s, d) for d in data])

    @staticmethod
    def __build_url(
            output: MimeoOutput,
    ) -> str:
        """Build a URL based on Mimeo Output Details.

        It populates a URL template using configured output details.
        When port is not configured - it uses host and endpoint values.
        Otherwise, host is followed by the colon and port.

        Parameters
        ----------
        output : MimeoOutput
            Configured Mimeo Output Details

        Returns
        -------
        str
            A valid URL address
        """
        if output.port is None:
            return f"{output.protocol}://{output.host}{output.endpoint}"
        return f"{output.protocol}://{output.host}:{output.port}{output.endpoint}"



"""The Mimeo Resources package.

This package contains all resources used in Mimeo, such as countries
and cities data or logging configuration file. It also includes
custom exceptions module:
* exc
    The Mimeo Resources Exceptions module.

To use this package, simply import the desired exception:
    from mimeo.resources.exc import ResourceNotFoundError
"""



"""The Mimeo Parser module.

It exports two parser classes:
    * MimeoArgumentParser
        A custom ArgumentParser for the Mimeo CLI.
    * MimeoConfigParser
        A class parsing source Mimeo Configuration with Mimeo
        arguments.
"""
from __future__ import annotations

import json
import logging
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import ClassVar

from mimeo import MimeoConfig
from mimeo.cli.exc import (EnvironmentNotFoundError,
                           EnvironmentsFileNotFoundError)
from mimeo.config import MimeoConfigFactory
from mimeo.config import constants as cc

logger = logging.getLogger(__name__)

DEFAULT_ENVS_PATH: str = "mimeo.envs.json"


class MimeoArgumentParser(ArgumentParser):
    """A custom ArgumentParser for the Mimeo CLI."""

    def __init__(
            self,
    ):
        """Initialize MimeoArgumentParser class.

        Extends ArgumentParser constructor with a Mimeo CLI details.

        It provides the following command line interface:

        usage: mimeo [OPTIONS] paths

        Generate data based on a template

        positional arguments:
          paths                 take paths to Mimeo Configuration files

        optional arguments:
          -h, --help            show this help message and exit
          -v, --version         show program's version number and exit

        Mimeo Configuration arguments:
          -F {xml,json}, --format {xml,json}
                                overwrite the output/format property
          -o {file,stdout,http}, --output {file,stdout,http}
                                overwrite the output/direction property
          -x {true,false}, --xml-declaration {true,false}
                                overwrite the output/xml_declaration property
          -i INDENT, --indent INDENT
                                overwrite the output/indent property
          -d DIRECTORY_PATH, --directory DIRECTORY_PATH
                                overwrite the output/directory_path property
          -f FILE_NAME, --file FILE_NAME
                                overwrite the output/file_name property
          -H HOST, --http-host HOST
                                overwrite the output/host property
          -p PORT, --http-port PORT
                                overwrite the output/port property
          -E ENDPOINT, --http-endpoint ENDPOINT
                                overwrite the output/endpoint property
          -U USERNAME, --http-user USERNAME
                                overwrite the output/username property
          -P PASSWORD, --http-password PASSWORD
                                overwrite the output/password property
          --http-method METHOD
                                overwrite the output/method property
          --http-protocol PROTOCOL
                                overwrite the output/protocol property
          -e ENVIRONMENT, --http-env ENVIRONMENT
                                overwrite the output http properties using a mimeo
                                env configuration
          --http-envs-file PATH
                                use a custom environments file
                                (by default: mimeo.envs.json)
          --raw
                                same as -o stdout
                                overwrite the output/direction property to stdout

        Logging arguments:
          --silent              disable INFO logs
          --debug               enable DEBUG mode
          --fine                enable FINE mode

        Other arguments:
          --sequentially        process Mimeo Configurations in a single thread
        """
        super().__init__(
            prog="mimeo",
            description="Generate data based on a template",
            usage="%(prog)s [OPTIONS] paths")
        self._add_positional_arguments()
        self._add_mimeo_configuration_arguments()
        self._add_logging_arguments()
        self._add_other_arguments()

    def _add_positional_arguments(
            self,
    ):
        """Add positional arguments."""
        self.add_argument(
            "-v",
            "--version",
            action="version",
            version="%(prog)s v1.1.0")
        self.add_argument(
            "paths",
            nargs="+",
            type=str,
            help="take paths to Mimeo Configuration files")

    def _add_mimeo_configuration_arguments(
            self,
    ):
        """Add arguments overwriting Mimeo Configuration."""
        mimeo_config_args = self.add_argument_group("Mimeo Configuration arguments")
        mimeo_config_args.add_argument(
            "-F",
            "--format",
            type=str,
            choices=["xml", "json"],
            help="overwrite the output/format property")
        mimeo_config_args.add_argument(
            "-o",
            "--output",
            type=str,
            choices=["file", "stdout", "http"],
            help="overwrite the output/direction property")
        mimeo_config_args.add_argument(
            "-x",
            "--xml-declaration",
            type=str,
            choices=["true", "false"],
            help="overwrite the output/xml_declaration property")
        mimeo_config_args.add_argument(
            "-i",
            "--indent",
            type=int,
            help="overwrite the output/indent property")
        mimeo_config_args.add_argument(
            "-d",
            "--directory",
            type=str,
            metavar="DIRECTORY_PATH",
            help="overwrite the output/directory_path property")
        mimeo_config_args.add_argument(
            "-f",
            "--file",
            type=str,
            metavar="FILE_NAME",
            help="overwrite the output/file_name property")
        mimeo_config_args.add_argument(
            "-H",
            "--http-host",
            type=str,
            metavar="HOST",
            help="overwrite the output/host property")
        mimeo_config_args.add_argument(
            "-p",
            "--http-port",
            type=str,
            metavar="PORT",
            help="overwrite the output/port property")
        mimeo_config_args.add_argument(
            "-E",
            "--http-endpoint",
            type=str,
            metavar="ENDPOINT",
            help="overwrite the output/endpoint property")
        mimeo_config_args.add_argument(
            "-U",
            "--http-user",
            type=str,
            metavar="USERNAME",
            help="overwrite the output/username property")
        mimeo_config_args.add_argument(
            "-P",
            "--http-password",
            type=str,
            metavar="PASSWORD",
            help="overwrite the output/password property")
        mimeo_config_args.add_argument(
            "--http-method",
            type=str,
            metavar="METHOD",
            help="overwrite the output/method property")
        mimeo_config_args.add_argument(
            "--http-protocol",
            type=str,
            metavar="PROTOCOL",
            help="overwrite the output/protocol property")
        mimeo_config_args.add_argument(
            "-e",
            "--http-env",
            type=str,
            metavar="ENVIRONMENT",
            help="overwrite the output http properties using a mimeo env configuration")
        mimeo_config_args.add_argument(
            "--http-envs-file",
            type=str,
            metavar="PATH",
            help=f"use a custom environments file (by default: {DEFAULT_ENVS_PATH})")
        mimeo_config_args.add_argument(
            "--raw",
            action="store_true",
            help="same as -o stdout - "
                 "overwrite the output/direction property to stdout")

    def _add_logging_arguments(
            self,
    ):
        """Add arguments customizing logs producing."""
        logging_args = self.add_argument_group("Logging arguments")
        logging_args_excl = logging_args.add_mutually_exclusive_group()
        logging_args_excl.add_argument(
            "--silent",
            action="store_true",
            help="disable INFO logs")
        logging_args_excl.add_argument(
            "--debug",
            action="store_true",
            help="enable DEBUG mode")
        logging_args_excl.add_argument(
            "--fine",
            action="store_true",
            help="enable FINE mode")

    def _add_other_arguments(
            self,
    ):
        """Add other arguments."""
        other_args = self.add_argument_group("Other arguments")
        other_args.add_argument(
            "--sequentially",
            action="store_true",
            help="process Mimeo Configurations in a single thread")


class MimeoConfigParser:
    """A class parsing source Mimeo Configuration with Mimeo arguments.

    Methods
    -------
    parse_config() -> MimeoConfig
        Parse a Mimeo Configuration using Mimeo arguments.
    """

    _ENVIRONMENT_PROPS: ClassVar[list] = [
        cc.OUTPUT_PROTOCOL_KEY,
        cc.OUTPUT_HOST_KEY,
        cc.OUTPUT_PORT_KEY,
        cc.OUTPUT_USERNAME_KEY,
        cc.OUTPUT_PASSWORD_KEY]

    _ENTRY_PATH_KEY: str = "entry_path"
    _GET_VALUE_KEY: str = "get_value"
    _ARGS_MAPPING: ClassVar[dict] = {
        "format": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_FORMAT_KEY],
        },
        "output": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_DIRECTION_KEY],
        },
        "xml_declaration": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_XML_DECLARATION_KEY],
            "get_value": lambda arg: arg.lower() == "true",
        },
        "indent": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_INDENT_KEY],
        },
        "directory": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_DIRECTORY_PATH_KEY],
        },
        "file": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_FILE_NAME_KEY],
        },
        "http_method": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_METHOD_KEY],
        },
        "http_protocol": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_PROTOCOL_KEY],
        },
        "http_host": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_HOST_KEY],
        },
        "http_port": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_PORT_KEY],
        },
        "http_endpoint": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_ENDPOINT_KEY],
        },
        "http_user": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_USERNAME_KEY],
        },
        "http_password": {
            "entry_path": [cc.OUTPUT_KEY, cc.OUTPUT_PASSWORD_KEY],
        },
    }

    def __init__(
            self,
            config_path: str,
            args: Namespace,
    ):
        """Initialize MimeoConfigParser class.

        Parameters
        ----------
        config_path: str
            A source config path
        args: Namespace
            Arguments parsed by MimeoArgumentParser
        """
        self._raw_config: dict = MimeoConfigFactory.parse_source(config_path)
        self._args: Namespace = args

    def parse_config(
            self,
    ) -> MimeoConfig:
        """Parse a Mimeo Configuration using Mimeo arguments.

        The parsing is made according to the following rule:
        configuration is overwritten from the most general to the most
        specific setting. It means that, whenever CLI arguments
        are overlapping (e.g. -e and -u), first the higher-level
        customization is applied (e.g. environment) and then the lower
        one (e.g. username).

        Returns
        -------
        mimeo_config : MimeoConfig
            A parsed Mimeo Configuration

        Raises
        ------
        EnvironmentsFileNotFoundError
            If environments file does not exist.
        EnvironmentNotFoundError
            If the http environment is not defined in the environments file
        """
        logger.info("Parsing Mimeo Configuration")
        config = dict(self._raw_config)
        config = self._parse_with_http_environment(config)
        config = self._parse_with_specific_args(config)
        mimeo_config = MimeoConfigFactory.parse(config)
        logger.fine("Parsed Mimeo Configuration: [%s]", mimeo_config)
        return mimeo_config

    def _parse_with_http_environment(
            self,
            config: dict,
    ) -> dict:
        """Parse Mimeo Configuration with an HTTP environment.

        When --http-envs-file argument has not been provided, then
        the default environments path is used: mimeo.envs.json.

        Parameters
        ----------
        config : dict
            A source Mimeo Configuration dictionary

        Returns
        -------
        config : dict
            An overwritten Mimeo Configuration dictionary when
            --http-env argument has been provided. Otherwise,
            the source one, without any modification.

        Raises
        ------
        EnvironmentsFileNotFoundError
            If environments file does not exist.
        EnvironmentNotFoundError
            If the http environment is not defined in the environments file
        """
        if self._args.http_env is None:
            return config

        if self._args.http_envs_file is None:
            env = self._get_environment(DEFAULT_ENVS_PATH, self._args.http_env)
        else:
            env = self._get_environment(self._args.http_envs_file, self._args.http_env)
        return self._overwrite_output_with_env(config, env)

    def _parse_with_specific_args(
            self,
            config: dict,
    ) -> dict:
        """Parse Mimeo Configuration with Mimeo Configuration args.

        This method uses an internal `dict` fetching each command line
        argument with a corresponding Mimeo Configuration entry.

        Parameters
        ----------
        config : dict
            A source Mimeo Configuration dictionary

        Returns
        -------
        config : dict
            An overwritten Mimeo Configuration dictionary when
            some configuration arguments have been provided. Otherwise,
            the source one, without any modification.
        """
        if self._args.raw:
            self._args.output = "stdout"
        for arg_name, mapping in self._ARGS_MAPPING.items():
            arg = getattr(self._args, arg_name, None)
            if arg is not None:
                entry_path = mapping.get(self._ENTRY_PATH_KEY, arg_name)
                get_value = mapping.get(self._GET_VALUE_KEY, lambda a: a)
                value = get_value(arg)
                config = self._overwrite_config_entry(config, entry_path, value)
        return config

    @classmethod
    def _overwrite_output_with_env(
            cls,
            config: dict,
            env: dict,
    ) -> dict:
        """Overwrite Mimeo Output Details with an environment config.

        Parameters
        ----------
        config : dict
            A source Mimeo Configuration dictionary
        env : dict
            An HTTP environment

        Returns
        -------
        config : dict
            An overwritten Mimeo Configuration.
        """
        for prop in cls._ENVIRONMENT_PROPS:
            value = env.get(prop)
            if value is not None:
                config_entry_path = [cc.OUTPUT_KEY, prop]
                config = cls._overwrite_config_entry(config, config_entry_path, value)
        return config

    @classmethod
    def _overwrite_config_entry(
            cls,
            config_entry: dict,
            entry_path: list,
            value: str | int | bool,
    ) -> dict:
        """Overwrite a Mimeo Configuration entry.

        Recursively finds an entry using `entry_path` list. If any of
        middle entries does not exist, it is initialized as an empty
        dictionary. Once the target entry is found, the `value` is put
        there.

        Parameters
        ----------
        config_entry : dict
            A Mimeo Configuration entry to get its direct child
        entry_path : list
            A list of Mimeo Configuration nested properties, e.g.
            ['output', 'direction'] points to the direction
            property under output details entry.
        value
            An overwriting value when `entry_path` contains only one
            element.

        Returns
        -------
        config_entry : dict
            An overwritten Mimeo Configuration entry.
        """
        direct_entry = entry_path[0]
        if len(entry_path) > 1:
            if direct_entry not in config_entry:
                config_entry[direct_entry] = {}
            value = cls._overwrite_config_entry(
                config_entry[direct_entry],
                entry_path[1:],
                value)
        else:
            logger.fine("Overwriting %s to [%s]", direct_entry, value)
        config_entry[direct_entry] = value
        return config_entry

    @staticmethod
    def _get_environment(
            envs_path: str,
            env_name: str,
    ) -> dict:
        """Load an HTTP environment from HTTP envs vile.

        Parameters
        ----------
        envs_path : str
            An HTTP envs file path
        env_name : str
            An HTTP environment name

        Returns
        -------
        env : dict
            An HTTP environment configuration

        Raises
        ------
        EnvironmentsFileNotFoundError
            If environments file does not exist.
        EnvironmentNotFoundError
            If the http environment is not defined in the environments file
        """
        if not Path(envs_path).exists():
            raise EnvironmentsFileNotFoundError(envs_path)

        with Path(envs_path).open() as envs_file:
            envs = json.load(envs_file)
            if env_name not in envs:
                raise EnvironmentNotFoundError(env_name, envs_path)

        env = envs[env_name]
        logger.debug("Using environment [%s] from file [%s]: [%s]",
                     env_name, envs_path, env)
        return env



"""The Mimeo Consumer Factory module.

It exports only one class:
    * ConsumerFactory
        A Factory class instantiating a Consumer based on Mimeo Config.
"""
from __future__ import annotations

from mimeo.config import constants as cc
from mimeo.config.exc import UnsupportedPropertyValueError
from mimeo.config.mimeo_config import MimeoConfig
from mimeo.consumers import Consumer, FileConsumer, HttpConsumer, RawConsumer


class ConsumerFactory:
    """A Factory class instantiating a Consumer based on Mimeo Config.

    Implementation of the Consumer class depends on the output direction configured.

    Attributes
    ----------
    FILE_DIRECTION
        The 'file' output direction
    STD_OUT_DIRECTION
        The 'stdout' output direction
    HTTP_DIRECTION
        The 'http' output direction

    Methods
    -------
    get_consumer(mimeo_config: MimeoConfig) -> Consumer
        Initialize a Consumer based on the Mimeo Output Direction.
    """

    FILE_DIRECTION: str = cc.OUTPUT_DIRECTION_FILE
    STD_OUT_DIRECTION: str = cc.OUTPUT_DIRECTION_STD_OUT
    HTTP_DIRECTION: str = cc.OUTPUT_DIRECTION_HTTP

    @staticmethod
    def get_consumer(
            mimeo_config: MimeoConfig,
    ) -> Consumer:
        """Initialize a Consumer based on the Mimeo Output Direction.

        Parameters
        ----------
        mimeo_config : MimeoConfig
            A Mimeo Configuration

        Returns
        -------
        Consumer
            A Consumer's implementation instance

        Raises
        ------
        UnsupportedPropertyValueError
            If the output direction is not supported
        """
        direction = mimeo_config.output.direction
        if direction == ConsumerFactory.STD_OUT_DIRECTION:
            return RawConsumer()
        if direction == ConsumerFactory.FILE_DIRECTION:
            return FileConsumer(mimeo_config.output)
        if direction == ConsumerFactory.HTTP_DIRECTION:
            return HttpConsumer(mimeo_config.output)
        raise UnsupportedPropertyValueError(
            cc.OUTPUT_DIRECTION_KEY,
            direction,
            cc.SUPPORTED_OUTPUT_DIRECTIONS)



"""The Mimeo Meta Exceptions module.

It contains all custom exceptions related to useful abstract classes:
    * InstanceNotAliveError
        A custom Exception class for using non-alive instance.
"""


from __future__ import annotations


class InstanceNotAliveError(Exception):
    """A custom Exception class for using non-alive instance.

    Raised when using OnlyOneAlive class that is not alive.
    """

    def __init__(
            self,
    ):
        """Initialize InstanceNotAliveError exception with details.

        Extends Exception constructor with a constant message.
        """
        super().__init__("The instance is not alive!")



"""The Mimeo JSON Generator module.

It exports only one class:
    * JSONGenerator
        A Generator implementation producing data in the JSON format.
"""
from __future__ import annotations

import json
import logging
from typing import Iterator

from mimeo.config import constants as cc
from mimeo.config.mimeo_config import MimeoConfig, MimeoTemplate
from mimeo.generators import Generator
from mimeo.utils import MimeoRenderer

logger = logging.getLogger(__name__)


class JSONGenerator(Generator):
    """A Generator implementation producing data in the JSON format.

    This Generator is instantiated for the 'json' output format
    and produces data using Mimeo Configuration.

    Methods
    -------
    generate(
        templates: list | Iterator[MimeoTemplate],
        parent: dict = None
    ) -> Iterator[dict]
        Generate JSON data based on the Mimeo Configuration.
    stringify(
        data: json
    ) -> str
        Stringify data generated by the generate() method.
    """

    def __init__(
            self,
            mimeo_config: MimeoConfig,
    ):
        """Initialize JSONGenerator class.

        Parameters
        ----------
        mimeo_config : MimeoConfig
            A Mimeo Configuration
        """
        self._indent: int = mimeo_config.output.indent

    @classmethod
    def generate(
            cls,
            templates: list | Iterator[MimeoTemplate],
            parent: dict | list | None = None,
    ) -> Iterator[dict]:
        """Generate JSON data based on the Mimeo Configuration.

        This function is used recursively when a Mimeo Configuration
        contains nested templates.
        It iterates through all templates configured and yields data
        units.

        Parameters
        ----------
        templates : list | Iterator[MimeoTemplate]
            A collection of Mimeo Templates to process
        parent : dict | list | None, default None
            A parent node for the currently processed template.
            It is passed only when a Mimeo Config contain nested
            templates.

        Returns
        -------
        Iterator[ElemTree.Element]
            Iterator for generated nodes
        """
        for template in templates:
            for data_unit in cls._process_single_template(template, parent):
                yield data_unit

    def stringify(
            self,
            data_unit: dict,
    ) -> str:
        """Stringify JSON data generated by the generate() method.

        Parameters
        ----------
        data_unit: dict
            A single data unit generated by the generate() method

        Returns
        -------
        str
            Stringified data unit
        """
        if self._indent is not None and self._indent != 0:
            return json.dumps(data_unit, indent=self._indent)

        return json.dumps(data_unit)

    @classmethod
    def _process_node(
            cls,
            parent: dict | list | None,
            node_meta: dict,
    ) -> dict | list:
        """Process a single template's node.

        Extends Generator's implementation by setting a parent node when it is None.
        It is required for JSON format to initialize an object.

        Parameters
        ----------
        parent : dict | list | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        dict | list
            A single data unit generated within a single template iteration.

        Raises
        ------
        InvalidSpecialFieldValueError
            If a special field value is dict or list
        SpecialFieldNotFoundError
            If a special field does not exist.
        """
        parent = parent if parent is not None else {}
        return super()._process_node(parent, node_meta)

    @classmethod
    def _pre_process_node(
            cls,
            node_meta: dict,
    ) -> dict:
        """Pre-process node's metadata.

        This function adjusts existing node's metadata and completes it with custom
        properties:
        * the name property is changed only for special fields
          - field name is extracted
        * the mimeo_util property is being initialized
          - True if element's value is a parametrized mimeo_util. Otherwise, False.
        * the special property is being initialized
          - True, when a field is special. Otherwise, False.

        Parameters
        ----------
        node_meta : dict
            Initial node's metadata

        Returns
        -------
        dict
            Complete node's metadata
        """
        name = node_meta["name"]
        value = node_meta["value"]
        is_mimeo_util = MimeoRenderer.is_parametrized_mimeo_util(value)
        is_special_field = name is not None and MimeoRenderer.is_special_field(name)
        if is_special_field:
            name = MimeoRenderer.get_special_field_name(name)

        return cls._node_meta(
            name=name,
            value=value,
            is_mimeo_util=is_mimeo_util,
            is_special_field=is_special_field)

    @classmethod
    def _process_complex_value(
            cls,
            parent: dict | list,
            node_meta: dict,
    ) -> dict | list:
        """Process a node with a complex value.

        The node is processed accordingly to its value type.
        When the type is dict, and it includes the _templates_ property, node
        is processed in a special way.

        Parameters
        ----------
        parent : dict | list
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        dict | list
            A processed node

        Raises
        ------
        InvalidSpecialFieldValueError
            If a special field value is dict or list
        SpecialFieldNotFoundError
            If a special field does not exist.
        """
        if (isinstance(node_meta["value"], dict) and
                cc.TEMPLATES_KEY not in node_meta["value"]):
            func = cls._process_dict_value
        elif isinstance(node_meta["value"], list):
            func = cls._process_list_value
        else:
            func = cls._process_templates_value
        return func(parent, node_meta)

    @classmethod
    def _process_dict_value(
            cls,
            parent: dict | list,
            node_meta: dict,
    ) -> dict | list:
        """Process a node with a dictionary value.

        It iterates through the dictionary items and processes each of them.

        Parameters
        ----------
        parent : dict | list
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        dict | list
            A processed node

        Raises
        ------
        InvalidSpecialFieldValueError
            If a special field value is dict or list
        SpecialFieldNotFoundError
            If a special field does not exist.

        Examples
        --------
        parent = {}
        node_meta = cls._node_meta(
            name="SomeField",
            value={"SomeChild1": 1, "SomeChild2": 2},
        )
        cls._process_dict_value(parent, node_meta)
        ->
        {
          "SomeField": {
            "SomeChild1": 1,
            "SomeChild2": 2,
          },
        }

        parent = []
        node_meta = cls._node_meta(
            name=None,
            value={"SomeChild1": 1, "SomeChild2": 2},
        )
        cls._process_dict_value(parent, node_meta)
        ->
        [
          {
            "SomeChild1": 1,
            "SomeChild2": 2,
          },
        ]
        """
        element = cls._create_node(parent, node_meta)
        for child_tag, child_value in node_meta["value"].items():
            cls._process_node(element, cls._node_meta(child_tag, child_value))
        return parent

    @classmethod
    def _process_list_value(
            cls,
            parent: dict | list,
            node_meta: dict,
    ) -> dict | list:
        """Process a node with a list value.

        It iterates through the list items and processes each of them: generates
        a child element for each value as direct children of the parent.

        Parameters
        ----------
        parent : dict | list
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        dict | list
            A processed node

        Raises
        ------
        InvalidSpecialFieldValueError
            If the special field value is dict or list
        SpecialFieldNotFoundError
            If the special field does not exist.

        Examples
        --------
        parent = {}
        node_meta = cls._node_meta(
            name="SomeField",
            value=[
                'value-1',
                {'SomeChild1': True, 'SomeChild2': False},
                {'_mimeo_util': {'_name': 'auto_increment', 'pattern': '{}'}}
            ],
        )
        cls._process_list_value_with_atomic_children(parent, node_meta)
        ->
        {
          "SomeField": [
            "value-1",
            {
                "SomeChild1": True,
                "SomeChild2": False,
            },
            1,
          ],
        }

        parent = []
        node_meta = cls._node_meta(
            name=None,
            value=[
                'value-1',
                {'SomeChild1': True, 'SomeChild2': False},
                {'_mimeo_util': {'_name': 'auto_increment', 'pattern': '{}'}}
            ],
        )
        cls._process_list_value_with_atomic_children(parent, node_meta)
        ->
        [
          [
            "value-1",
            {
                "SomeChild1": True,
                "SomeChild2": False,
            },
            1,
          ],
        ]
        """
        element = cls._create_node(parent, node_meta)
        for child in node_meta["value"]:
            node_meta = cls._node_meta(None, child)
            cls._process_node(element, node_meta)
        return parent

    @classmethod
    def _process_templates_value(
            cls,
            parent: dict | list,
            node_meta: dict,
    ) -> dict | list:
        """Process a node with a dictionary value storing templates.

        It iterates through the templates and generates data based on them.
        If parent is a dict, the property name will take a value of an array.
        All properties at the same level as _templates_ will be ignored.

        Parameters
        ----------
        parent : dict | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        dict | list
            A processed node

        Raises
        ------
        InvalidSpecialFieldValueError
            If a special field value is dict or list
        SpecialFieldNotFoundError
            If a special field does not exist.

        Examples
        --------
        parent = []
        node_meta = cls._node_meta(
            name=None,
            value={"_templates_": [
                {
                  "count": 10,
                  "model": {
                    "SomeChild": {
                      "Node1": 1,
                      "Node2": "value-2",
                      "Node3": true
                    }
                  }
                }
            ]},
        )
        cls._process_templates_value(parent, node_meta)
        ->
        [
          {
            "SomeChild": {
              "Node1": 1,
              "Node2": "value-2",
              "Node3": true,
            },
          },
          {
            "SomeChild": {
              "Node1": 1,
              "Node2": "value-2",
              "Node3": true,
            },
          },
          ... x10
        ]

        parent = {}
        node_meta = cls._node_meta(
            name="SomeField",
            value={"_templates_": [
                {
                  "count": 10,
                  "model": {
                    "SomeChild": {
                      "Node1": 1,
                      "Node2": "value-2",
                      "Node3": true
                    }
                  }
                }
            ]},
        )
        cls._process_templates_value(parent, node_meta)
        ->
        {
          "SomeField": [
            {
              "SomeChild": {
                "Node1": 1,
                "Node2": "value-2",
                "Node3": true,
              },
            },
            {
              "SomeChild": {
                "Node1": 1,
                "Node2": "value-2",
                "Node3": true,
              },
            },
            ... x10
          ],
        }
        """
        templates = (MimeoTemplate(template)
                     for template in node_meta["value"][cc.TEMPLATES_KEY])
        target = parent
        if isinstance(parent, dict):
            parent[node_meta["name"]] = []
            target = parent[node_meta["name"]]

        target.extend(cls.generate(templates))
        return parent

    @classmethod
    def _process_atomic_value(
            cls,
            parent: dict | list,
            node_meta: dict,
    ) -> dict | list:
        """Process a node with an atomic value.

        A parametrized Mimeo Util is considered as an atomic value as representing one.
        It renders a value for the node.

        Parameters
        ----------
        parent : dict | list
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        dict | list
            A processed node

        Raises
        ------
        InvalidSpecialFieldValueError
            If a special field value is dict or list
        SpecialFieldNotFoundError
            If a special field does not exist.

        Examples
        --------
        parent = {}
        node_meta = cls._node_meta(
            name="SomeField",
            value="value-1",
        )
        cls._process_atomic_value(parent, node_meta)
        ->
        {
          "SomeField": "value-1"
        }

        parent = []
        node_meta = cls._node_meta(
            name=None,
            value="value-1",
        )
        cls._process_atomic_value(parent, node_meta)
        ->
        [
          "value-1"
        ]
        """
        value = super()._render_atomic_value(node_meta)
        if isinstance(parent, dict):
            parent[node_meta["name"]] = value
        elif isinstance(parent, list):
            parent.append(value)
        logger.fine("Rendered value [%s]", value)
        return parent

    @staticmethod
    def _create_node(
            parent: dict | list,
            node_meta: dict,
    ) -> dict | list:
        """Create an JSON element based on the `parent` and entry value types.

        Parameters
        ----------
        parent : dict | list
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        dict | list
            If the new node's value is dict, returns dict. Otherwise, a list.

        Examples
        --------
        parent = {}
        node_meta = cls._node_meta(
            name="SomeEntity",
            value={"SomeField": "value-1"},
        )
        node = cls._create_node(parent, node_meta)
        ->
        parent: {"SomeEntity": {}}
        node: {}

        parent = []
        node_meta = cls._node_meta(
            name=None,
            value={"SomeField": "value-1"},
        )
        node = cls._create_node(parent, node_meta)
        ->
        parent: [{}]
        node: {}

        parent = []
        node_meta = cls._node_meta(
            name=None,
            value=["value-1"],
        )
        node = cls._create_node(parent, node_meta)
        ->
        parent: [[]]
        node: []

        parent = {}
        node_meta = cls._node_meta(
            name="SomeEntity",
            value=["value-1"],
        )
        node = cls._create_node(parent, node_meta)
        ->
        parent: {"SomeEntity": []}
        node: []
        """
        new_node = {} if isinstance(node_meta["value"], dict) else []
        if isinstance(parent, list):
            parent.append(new_node)
            return parent[-1]

        parent[node_meta["name"]] = new_node
        return parent[node_meta["name"]]



"""The Mimeo Consumers package.

It contains modules supporting the Mimeo Config output directions:
* consumer
    The Mimeo Consumer module.
* consumer_factory
    The Mimeo Consumer Factory module.
* raw_consumer
    The Mimeo Raw Consumer module.
* file_consumer
    The Mimeo File Consumer module.
* http_consumer
    The Mimeo HTTP Consumer module.

This package exports the following classes:
* Consumer:
    An abstract class for data consumers in Mimeo.
* ConsumerFactory:
    A Factory class instantiating a Consumer based on Mimeo Config.
* RawConsumer:
    A Consumer implementation printing data in the standard output.
    Corresponds to the 'stdout' output direction
* FileConsumer:
    A Consumer implementation saving data in the filesystem.
    Corresponds to the 'file' output direction
* HttpConsumer:
    A Consumer implementation sending data in HTTP requests.
    Corresponds to the 'http' output direction

To use this package, simply import the desired class:
    from mimeo.consumers import ConsumerFactory
"""

from .consumer import Consumer
from .file_consumer import FileConsumer
from .raw_consumer import RawConsumer
from .http_consumer import HttpConsumer
from .consumer_factory import ConsumerFactory

__all__ = ["Consumer", "FileConsumer", "RawConsumer", "HttpConsumer", "ConsumerFactory"]



"""The Mimeo XML Generator module.

It exports only one class:
    * XMLGenerator
        A Generator implementation producing data in the XML format.
"""
from __future__ import annotations

import logging
import xml.etree.ElementTree as ElemTree
from typing import Iterator
from xml.dom import minidom

from mimeo.config import constants as cc
from mimeo.config.mimeo_config import MimeoConfig, MimeoTemplate
from mimeo.generators import Generator
from mimeo.generators.exc import UnsupportedStructureError
from mimeo.utils import MimeoRenderer

logger = logging.getLogger(__name__)


class XMLGenerator(Generator):
    """A Generator implementation producing data in the XML format.

    This Generator is instantiated for the 'xml' output format
    and produces data using Mimeo Configuration.

    Methods
    -------
    generate(
        templates: list | Iterator[MimeoTemplate],
        parent: ElemTree.Element = None
    ) -> Iterator[ElemTree.Element]
        Generate XML data based on the Mimeo Configuration.
    stringify(
        data_unit: ElemTree.Element
    ) -> str
        Stringify data generated by the generate() method.
    """

    def __init__(
            self,
            mimeo_config: MimeoConfig,
    ):
        """Initialize XMLGenerator class.

        Parameters
        ----------
        mimeo_config : MimeoConfig
            A Mimeo Configuration
        """
        self._indent: int = mimeo_config.output.indent
        self._xml_declaration: bool = mimeo_config.output.xml_declaration

    @classmethod
    def generate(
            cls,
            templates: list | Iterator[MimeoTemplate],
            parent: ElemTree.Element | None = None,
    ) -> Iterator[ElemTree.Element]:
        """Generate XML data based on the Mimeo Configuration.

        This function is used recursively when a Mimeo Configuration
        contains nested templates.
        It iterates through all templates configured and yields data
        units.

        Parameters
        ----------
        templates : list | Iterator[MimeoTemplate]
            A collection of Mimeo Templates to process
        parent : ElemTree.Element, default None
            A parent XML node for the currently processed template.
            It is passed only when a Mimeo Config contain nested
            templates.

        Returns
        -------
        Iterator[ElemTree.Element]
            Iterator for generated nodes
        """
        for template in templates:
            for data_unit in cls._process_single_template(template, parent):
                yield data_unit

    def stringify(
            self,
            data_unit: ElemTree.Element,
    ) -> str:
        """Stringify XML data generated by the generate() method.

        Parameters
        ----------
        data_unit: ElemTree.Element
            A single data unit generated by the generate() method

        Returns
        -------
        str
            Stringified data unit
        """
        if self._indent is None or self._indent == 0:
            node_str = ElemTree.tostring(
                data_unit,
                encoding="utf-8",
                method="xml",
                xml_declaration=self._xml_declaration)
        else:
            xml_string = ElemTree.tostring(data_unit)
            xml_minidom = minidom.parseString(xml_string)
            if self._xml_declaration is False:
                xml_minidom = xml_minidom.childNodes[0]
            node_str = xml_minidom.toprettyxml(
                indent=" " * self._indent,
                encoding="utf-8")
        return node_str.decode("ascii")

    @classmethod
    def _pre_process_node(
            cls,
            node_meta: dict,
    ) -> dict:
        """Pre-process node's metadata.

        This function adjusts existing node's metadata and completes it with custom
        properties:
        * the name property is changed only for special fields
          - field name is extracted
        * the value property is being modified for dicts including attributes
          - properties starting with '@' are being removed from the dict
        * the attrs property
          - is being populated by all properties starting with '@'
          - takes the default value (an empty dict) when there's no such properties
        * the mimeo_util property is being initialized
          - True if element's value is a parametrized mimeo_util. Otherwise, False.
        * the special property is being initialized
          - True, when a field is special. Otherwise, False.

        Parameters
        ----------
        node_meta : dict
            Initial node's metadata

        Returns
        -------
        dict
            Complete node's metadata
        """
        name = node_meta["name"]
        value = node_meta["value"]
        attrs = node_meta["attrs"] if node_meta["attrs"] is not None else {}
        is_mimeo_util = MimeoRenderer.is_parametrized_mimeo_util(value)
        is_special_field = MimeoRenderer.is_special_field(name)
        if is_special_field:
            name = MimeoRenderer.get_special_field_name(name)
        if isinstance(value, dict):
            value_copy = dict(value)
            for prop in value:
                if prop.startswith(cc.MODEL_ATTRIBUTES_KEY):
                    attrs[prop[1:]] = value_copy.pop(prop)
            value = value.get(cc.MODEL_TEXT_VALUE_KEY, value_copy)

        return cls._node_meta(
            name,
            value,
            attrs,
            is_mimeo_util,
            is_special_field)

    @classmethod
    def _process_complex_value(
            cls,
            parent: ElemTree.Element | None,
            node_meta: dict,
    ) -> ElemTree.Element | None:
        """Process a node with a complex value.

        The node is processed accordingly to its value type.
        When the type is list and name is _templates_, node is processed
        in a special way.

        Parameters
        ----------
        parent : ElemTree.Element | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        ElemTree.Element
            A processed node

        Raises
        ------
        UnsupportedStructureError
            If a list value elements are not atomic-only or dict-only.
        InvalidSpecialFieldValueError
            If a special field value is dict or list
        SpecialFieldNotFoundError
            If a special field does not exist.
        """
        if isinstance(node_meta["value"], dict):
            func = cls._process_dict_value
        elif (isinstance(node_meta["value"], list) and
              node_meta["name"] != cc.TEMPLATES_KEY):
            func = cls._process_list_value
        else:
            func = cls._process_templates_value
        return func(parent, node_meta)

    @classmethod
    def _process_dict_value(
            cls,
            parent: ElemTree.Element | None,
            node_meta: dict,
    ) -> ElemTree.Element:
        """Process a node with a dictionary value.

        It iterates through the dictionary items and processes each of them.

        Parameters
        ----------
        parent : ElemTree.Element | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        ElemTree.Element
            A processed node

        Raises
        ------
        UnsupportedStructureError
            If a list value elements are not atomic-only or dict-only.
        InvalidSpecialFieldValueError
            If a special field value is dict or list
        SpecialFieldNotFoundError
            If a special field does not exist.

        Examples
        --------
        parent = ElemTree.Element("Root")
        node_meta = cls._node_meta(
            name="SomeField",
            value={"SomeChild1": 1, "SomeChild2": 2},
        )
        cls._process_dict_value(parent, node_meta)
        ->
        <SomeField>
            <SomeChild1>1</SomeChild1>
            <SomeChild2>2</SomeChild2>
        </SomeField>
        """
        element = cls._create_node(parent, node_meta)
        for child_tag, child_value in node_meta["value"].items():
            cls._process_node(element, cls._node_meta(child_tag, child_value))
        return element

    @classmethod
    def _process_list_value(
            cls,
            parent: ElemTree.Element,
            node_meta: dict,
    ) -> ElemTree.Element:
        """Process a node with a list value.

        It iterates through the list items and processes each of them: generates
        a child element for each value as direct children of the parent.

        Parameters
        ----------
        parent : ElemTree.Element | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        ElemTree.Element
            A processed node

        Raises
        ------
        UnsupportedStructureError
            If any of the list value element is a list.
        InvalidSpecialFieldValueError
            If the special field value is dict or list
        SpecialFieldNotFoundError
            If the special field does not exist.

        Examples
        --------
        parent = ElemTree.Element("Root")
        node_meta = cls._node_meta(
            name="SomeField",
            value=[
                'value-1',
                {'SomeChild1': True, 'SomeChild2': False},
                {'_mimeo_util': {'_name': 'auto_increment', 'pattern': '{}'}}
            ],
        )
        cls._process_list_value_with_atomic_children(parent, node_meta)
        ->
        <Root>
            <SomeField>value-1</SomeField>
            <SomeField>
                <SomeChild1>true</SomeChild1>
                <SomeChild2>false</SomeChild2>
            </SomeField>
            <SomeField>1</SomeField>
        </Root>
        """
        for child in node_meta["value"]:
            if isinstance(child, list):
                raise UnsupportedStructureError(
                    node_meta["name"],
                    node_meta["value"])
            node_meta = cls._node_meta(
                name=node_meta["name"],
                value=child,
                attrs=node_meta["attrs"])
            cls._process_node(parent, node_meta)
        return parent

    @classmethod
    def _process_templates_value(
            cls,
            parent: ElemTree.Element,
            node_meta: dict,
    ) -> ElemTree.Element:
        """Process a node with a dictionary value storing templates.

        It iterates through the templates and generates data based on them.

        Parameters
        ----------
        parent : ElemTree.Element | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        ElemTree.Element
            A processed node

        Raises
        ------
        UnsupportedStructureError
            If a list value elements are not atomic-only or dict-only.
        InvalidSpecialFieldValueError
            If a special field value is dict or list
        SpecialFieldNotFoundError
            If a special field does not exist.

        Examples
        --------
        parent = ElemTree.Element("Root")
        node_meta = cls._node_meta(
            name="SomeField",
            value={"_templates_": [
                {
                  "count": 10,
                  "model": {
                    "SomeChild": {
                      "Node1": 1,
                      "Node2": "value-2",
                      "Node3": true
                    }
                  }
                }
            ]},
        )
        cls._process_templates_value(parent, node_meta)
        ->
        <Root>
            <SomeField>
                <SomeChild><Node1>1</Node1><Node2>value-2</Node2><Node3>true</Node3></SomeChild>
                <SomeChild><Node1>1</Node1><Node2>value-2</Node2><Node3>true</Node3></SomeChild>
                ... x10
            </SomeField>
        </Root>
        """
        templates = (MimeoTemplate(template) for template in node_meta["value"])
        for _ in cls.generate(templates, parent):
            pass
        return parent

    @classmethod
    def _process_atomic_value(
            cls,
            parent: ElemTree.Element,
            node_meta: dict,
    ) -> ElemTree.Element:
        """Process a node with an atomic value.

        A parametrized Mimeo Util is considered as an atomic value as representing one.
        It renders a value for the node.

        Parameters
        ----------
        parent : ElemTree.Element | None
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        ElemTree.Element
            A processed node

        Raises
        ------
        UnsupportedStructureError
            If a list value elements are not atomic-only or dict-only.
        InvalidSpecialFieldValueError
            If a special field value is dict or list
        SpecialFieldNotFoundError
            If a special field does not exist.

        Examples
        --------
        parent = ElemTree.Element("Root")
        node_meta = cls._node_meta(
            name="SomeField",
            value="value-1",
        )
        cls._process_atomic_value(parent, node_meta)
        ->
        <SomeField>value-1</SomeField>
        """
        element = cls._create_node(parent, node_meta)
        value = super()._render_atomic_value(node_meta)
        val_str = str(value) if value is not None else ""
        element.text = val_str.lower() if isinstance(value, bool) else val_str
        logger.fine("Rendered value [%s]", element.text)
        return element

    @staticmethod
    def _create_node(
            parent: ElemTree.Element,
            node_meta: dict,
    ) -> ElemTree.Element | ElemTree.SubElement:
        """Create an XML node based on the `parent` existence.

        Parameters
        ----------
        parent : ElemTree.Element
            A parent node
        node_meta : dict
            Node's metadata

        Returns
        -------
        ElemTree.Element | ElemTree.SubElement
            If the `parent` is None, returns ElemTree.Element.
            Otherwise, returns ElemTree.SubElement
        """
        if parent is None:
            return ElemTree.Element(
                node_meta["name"],
                attrib=node_meta["attrs"])
        return ElemTree.SubElement(
            parent,
            node_meta["name"],
            attrib=node_meta["attrs"],
        )



"""The Mimeo Job module.

It exports a single class:
    * MimeoJob
        A class representing a single Mimeo processing job.
"""
from __future__ import annotations

import logging
from argparse import Namespace
from os import walk
from pathlib import Path

from mimeo import MimeoConfig, Mimeograph
from mimeo.cli import MimeoArgumentParser, MimeoConfigParser
from mimeo.cli.exc import PathNotFoundError

logger = logging.getLogger(__name__)


class MimeoJob:
    """A class representing a single Mimeo processing job.

    It is meant to be used in context of a command line. First it
    parses command line arguments using MimeoArgumentParser and then
    prepares logging. In the next step each Mimeo Configuration is
    parsed by MimeoConfigParser and provided arguments. Having prepared
    Mimeo Configuration, data processing starts.

    Methods
    -------
    run()
        Executes a Mimeo Job based on the CLI arguments.
    """

    def __init__(
            self,
    ):
        """Initialize MimeoJob class."""
        self._args: Namespace = MimeoArgumentParser().parse_args()

    def run(
            self,
    ):
        """Execute a Mimeo Job based on the CLI arguments.

        First it customizes a log level. After that all Mimeo Configs paths are
        collected. Each of them is used in the next steps, which are: (1) parsing
        the configuration and (2) processing it. Depending on arguments provided,
        it processes configs sequentially or in parallel.
        """
        self._customize_log_level(self._args)
        logger.info("Starting a Mimeo job")
        if self._args.sequentially:
            self._process_sequentially()
        else:
            self._process_in_parallel()

    @staticmethod
    def _customize_log_level(
            args,
    ):
        """Customize the log level based on command line arguments."""
        if args.silent:
            logging.getLogger("mimeo").setLevel(logging.WARNING)
        elif args.debug:
            logging.getLogger("mimeo").setLevel(logging.DEBUG)
        elif args.fine and hasattr(logging, "FINE"):
            logging.getLogger("mimeo").setLevel(logging.FINE)

    def _process_sequentially(self):
        config_paths = self._get_config_paths(self._args.paths)
        for config_path in config_paths:
            mimeo_config = self._get_mimeo_config(config_path, self._args)
            Mimeograph.process(mimeo_config)

    def _process_in_parallel(self):
        config_paths = self._get_config_paths(self._args.paths)
        with Mimeograph() as mimeo:
            for config_path in config_paths:
                mimeo_config = self._get_mimeo_config(config_path, self._args)
                mimeo.submit((config_path, mimeo_config))

    @staticmethod
    def _get_config_paths(
            paths: list[str],
    ) -> list[str]:
        """Collect Mimeo Configuration paths.

        This method traverses directory paths and collects all files
        within.

        Parameters
        ----------
        paths : list[str]
            A list of paths provided in command line

        Returns
        -------
        file_paths : list[str]
            A list of file paths
        """
        file_paths = []
        for file_path in paths:
            if Path(file_path).is_dir():
                for dir_path, _, file_names in walk(file_path):
                    for file_name in file_names:
                        path = Path(f"{dir_path}/{file_name}")
                        file_paths.append(str(path))
            elif Path(file_path).is_file():
                path = Path(file_path)
                file_paths.append(str(path))
            else:
                raise PathNotFoundError(file_path)
        return file_paths

    @classmethod
    def _get_mimeo_config(
            cls,
            config_path: str,
            args: Namespace,
    ) -> MimeoConfig:
        """Return parsed Mimeo Configuration.

        This method parses a raw configuration with command line
        arguments using a MimeoConfigParser instance.

        Parameters
        ----------
        config_path : str
            A raw configuration path
        args
            Command line arguments parsed by MimeoArgumentParser

        Returns
        -------
        MimeoConfig
            A parsed Mimeo Configuration

        Raises
        ------
        EnvironmentsFileNotFoundError
            If environments file does not exist.
        EnvironmentNotFoundError
            If the http environment is not defined in the environments file
        """
        mimeo_config_parser = MimeoConfigParser(config_path, args)
        return mimeo_config_parser.parse_config()



"""The Countries module.

It exports classes related to currencies CSV data:
    * Currency
        DTO class representing a single row in currencies CSV data.
    * CurrenciesDB
        Class exposing READ operations on currencies CSV data.
"""
from __future__ import annotations

from typing import ClassVar

import pandas
from pandas import DataFrame

from mimeo import tools
from mimeo.database.exc import InvalidIndexError


class Currency:
    """DTO class representing a single row in currencies CSV data.

    Attributes
    ----------
    code : str
        A currency code
    name : str
        A currency name
    countries : list[str]
        A countries using the currency
    """

    def __init__(
            self,
            code: str,
            name: str,
            countries: list[str],
    ):
        """Initialize Currency class.

        Parameters
        ----------
        code : str
            A currency code
        name : str
            A currency name
        countries : list
            A countries using the currency
        """
        self.code: str = code
        self.name: str = name
        self.countries: list[str] = countries

    def __str__(
            self,
    ) -> str:
        """Stringify the Currency instance.

        Returns
        -------
        str
            A stringified `dict` representation of the Currency instance
        """
        return str({
            "code": self.code,
            "name": self.name,
            "countries": self.countries,
        })

    def __repr__(
            self,
    ) -> str:
        """Represent the Currency instance.

        Returns
        -------
        str
            A python representation of the Currency instance
        """
        return (f"Currency("
                f"code='{self.code}', "
                f"name='{self.name}', "
                f"countries={self.countries})")


class CurrenciesDB:
    """Class exposing READ operations on currencies CSV data.

    Attributes
    ----------
    NUM_OF_RECORDS : int
        A number of rows in currencies CSV data

    Methods
    -------
    get_currencies() -> list[Currency]
        Get all currencies.
    get_currency_of(country_name: str) -> Currency | None
        Get currency of a specific country.
    get_currency_at(index: int) -> Currency
        Get a currency at `index` position.
    """

    NUM_OF_RECORDS: int = 169
    _CURRENCIES_DB: str = "currencies.csv"
    _CURRENCIES_DF: DataFrame = None
    _CURRENCIES: ClassVar[list] = None
    _COUNTRY_CURRENCIES: ClassVar[dict] = {}

    def get_currency_at(
            self,
            index: int,
    ) -> Currency:
        """Get a currency at `index` position.

        Parameters
        ----------
        index : int
            A currency row index

        Returns
        -------
        Currency
            A specific currency

        Raises
        ------
        InvalidIndexError
            If the provided `index` is out of bounds
        """
        currencies = CurrenciesDB._get_currencies()
        try:
            return currencies[index]
        except IndexError:
            last_index = CurrenciesDB.NUM_OF_RECORDS-1
            raise InvalidIndexError(index, last_index) from IndexError

    def get_currency_of(
            self,
            country_name: str,
    ) -> Currency | None:
        """Get a currency of a specific country.

        Parameters
        ----------
        country_name : str
            A country name to find a currency

        Returns
        -------
        Currency | None
            A currency used in a specific country or None
        """
        return CurrenciesDB._get_country_currency(country_name)

    def get_currencies(
            self,
    ) -> list[Currency]:
        """Get all currencies.

        Returns
        -------
        list[Currency]
            List of all currencies
        """
        return CurrenciesDB._get_currencies().copy()

    @classmethod
    def _get_country_currency(
            cls,
            country_name: str,
    ) -> Currency | None:
        """Get currency of a specific country from cache.

        The country's currency is initialized for the first time and cached in internal
        class attribute.

        Parameters
        ----------
        country_name : str
            A country ISO3 code to filter currencies

        Returns
        -------
        Currency | None
            A currency used in a specific country or None
        """
        if country_name not in cls._COUNTRY_CURRENCIES:
            currencies = cls._get_currencies()
            country_currency = filter(lambda c: country_name in c.countries, currencies)
            cls._COUNTRY_CURRENCIES[country_name] = next(country_currency, None)
        return cls._COUNTRY_CURRENCIES[country_name]

    @classmethod
    def _get_currencies(
            cls,
    ) -> list[Currency]:
        """Get all currencies from cache.

        The currencies list is initialized for the first time and cached in internal
        class attribute.

        Returns
        -------
        list[Currency]
            List of all currencies
        """
        if cls._CURRENCIES is None:
            cls._CURRENCIES = [Currency(row.CODE, row.NAME, row.COUNTRIES)
                               for row in cls._get_currencies_df().itertuples()]
        return cls._CURRENCIES

    @classmethod
    def _get_currencies_df(
            cls,
    ) -> pandas.DataFrame:
        """Load currencies CSV data and save in internal class attribute."""
        if cls._CURRENCIES_DF is None:
            data = tools.get_resource(cls._CURRENCIES_DB)
            cls._CURRENCIES_DF = pandas.read_csv(data)
        return cls._CURRENCIES_DF



"""The Mimeo Resources Exceptions module.

It contains all custom exceptions related to Mimeo resources:
    * ResourceNotFoundError
        A custom Exception class for a not found resource.
"""


from __future__ import annotations


class ResourceNotFoundError(Exception):
    """A custom Exception class for a not found resource.

    Raised while attempting to get a resource that does exist.
    """

    def __init__(
            self,
            resource_name: str,
    ):
        """Initialize ResourceNotFoundError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        resource_name : str
            A resource name
        """
        super().__init__(f"No such resource: [{resource_name}]")



"""The Mimeo Configuration Exceptions module.

It contains all custom exceptions related to Mimeo Configuration:
    * UnsupportedPropertyValueError
        A custom Exception class for unsupported properties' values.
    * MissingRequiredPropertyError
        A custom Exception class for missing required properties.
    * InvalidIndentError
        A custom Exception class for invalid indent configuration.
    * InvalidVarsError
        A custom Exception class for invalid vars' configuration.
    * InvalidRefsError
        A custom Exception class for invalid refs' configuration.
    * InvalidVarsError.Code
        An Enumeration class for InvalidVarsError error codes.
    * InvalidMimeoModelError
        A custom Exception class for invalid model configuration.
    * InvalidMimeoModelError.Code
        An Enumeration class for InvalidMimeoModelError error codes.
    * InvalidMimeoTemplateError
        A custom Exception class for invalid template configuration.
    * InvalidMimeoConfigError
        A custom Exception class for invalid mimeo configuration.
    * InvalidMimeoConfigError.Code
        An Enumeration class for InvalidMimeoConfigError error codes.
    * UnsupportedMimeoConfigSourceError
        A custom Exception class for unsupported Mimeo Configuration source.
    * MimeoConfigurationNotFoundError
        A custom Exception class for no mimeo_configuration node in source xml.
"""


from __future__ import annotations

from enum import Enum
from typing import Any


class UnsupportedPropertyValueError(Exception):
    """A custom Exception class for unsupported properties' values.

    Raised when a Mimeo Configuration property points to a value
    not being supported by Mimeo.
    """

    def __init__(
            self,
            prop: str,
            val: str,
            supported_values: tuple,
    ):
        """Initialize UnsupportedPropertyValueError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        prop : str
            A property name
        val : str
            A property value
        supported_values : tuple
            A list of supported values for the property
        """
        super().__init__(f"Provided {prop} [{val}] is not supported! "
                         f"Supported values: [{', '.join(supported_values)}].")


class MissingRequiredPropertyError(Exception):
    """A custom Exception class for missing required properties.

    Raised when a Mimeo Configuration does not contain a required
    property.
    """

    def __init__(
            self,
            details: list,
    ):
        """Initialize MissingRequiredPropertyError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        details : list
            Missing details
        """
        details_str = ", ".join(details)
        super().__init__(f"Missing required fields in HTTP output details: "
                         f"{details_str}")


class InvalidIndentError(Exception):
    """A custom Exception class for invalid indent configuration.

    Raised when a configured indent is negative.
    """

    def __init__(
            self,
            indent: int,
    ):
        """Initialize InvalidIndentError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        indent : int
            A configured indent
        """
        super().__init__(f"Provided indent [{indent}] is negative!")


class InvalidVarsError(Exception):
    """A custom Exception class for invalid vars' configuration.

    Raised when vars are not configured properly.
    """

    class Code(Enum):
        """An Enumeration class for InvalidVarsError error codes.

        Attributes
        ----------
        ERR_1: str
            An error code for vars not being a dictionary
        ERR_2: str
            An error code for vars with non-atomic values
        ERR_3: str
            An error code for vars with not allowed characters
        """

        ERR_1: str = "NOT_A_DICT"
        ERR_2: str = "COMPLEX_VALUE"
        ERR_3: str = "INVALID_NAME"

    def __init__(
            self,
            code: InvalidVarsError.Code,
            **kwargs,
    ):
        """Initialize InvalidVarsError exception with details.

        Extends Exception constructor with a custom message. The message depends on
        an internal InvalidVarsError code.

        Parameters
        ----------
        code : InvalidVarsError.Code
            An internal error code
        kwargs
            An error details
        """
        msg = self._get_msg(code, kwargs)
        super().__init__(msg)

    @classmethod
    def _get_msg(
            cls,
            code: InvalidVarsError.Code,
            details: dict,
    ):
        """Return a custom message based on an error code.

        Parameters
        ----------
        code : InvalidVarsError.Code
            An internal error code
        details : dict
            An error details

        Returns
        -------
        str
            A custom error message

        Raises
        ------
        ValueError
            If the code argument is not InvalidVarsError.Code enum
        """
        if code == cls.Code.ERR_1:
            return f"vars property does not store an object: {details['vars']}"
        if code == cls.Code.ERR_2:
            return (f"Provided var [{details['var']}] is invalid "
                    f"(you can use ony atomic values and Mimeo Utils)!")
        if code == cls.Code.ERR_3:
            return (f"Provided var [{details['var']}] is invalid "
                    f"(you can use upper-cased name with underscore and digits, "
                    f"starting with a letter)!")

        msg = f"Provided error code is not a {cls.__name__}.Code enum!"
        raise ValueError(msg)


class InvalidRefsError(Exception):
    """A custom Exception class for invalid refs' configuration.

    Raised when vars are not configured properly.
    """

    class Code(Enum):
        """An Enumeration class for InvalidRefsError error codes.

        Attributes
        ----------
        ERR_1: str
            An error code for refs not being a dictionary
        ERR_2: str
            An error code for a ref not being a dictionary
        ERR_3: str
            An error code for refs with missing required details
        ERR_4: str
            An error code for refs having same name as a Mimeo Util or a Mimeo Var
        """

        ERR_1: str = "REFS_NOT_A_DICT"
        ERR_2: str = "REF_NOT_A_DICT"
        ERR_3: str = "MISSING_DETAILS"
        ERR_4: str = "FORBIDDEN_NAME"

    def __init__(
            self,
            code: InvalidRefsError.Code,
            **kwargs,
    ):
        """Initialize InvalidRefsError exception with details.

        Extends Exception constructor with a custom message. The message depends on
        an internal InvalidRefsError code.

        Parameters
        ----------
        code : InvalidRefsError.Code
            An internal error code
        kwargs
            An error details
        """
        msg = self._get_msg(code, kwargs)
        super().__init__(msg)

    @classmethod
    def _get_msg(
            cls,
            code: InvalidVarsError.Code,
            details: dict,
    ):
        """Return a custom message based on an error code.

        Parameters
        ----------
        code : InvalidRefsError.Code
            An internal error code
        details : dict
            An error details

        Returns
        -------
        str
            A custom error message

        Raises
        ------
        ValueError
            If the code argument is not InvalidRefsError.Code enum
        """
        if code == cls.Code.ERR_1:
            return f"refs property does not store an object: {details['refs']}"
        if code == cls.Code.ERR_2:
            return f"The following ref does not store an object: {details['ref']}"
        if code == cls.Code.ERR_3:
            required_details = ", ".join(details["required"])
            references = ", ".join(details["refs"])
            return (f"Missing required details [{required_details}] in the following "
                    f"refs [{references}]")
        if code == cls.Code.ERR_4:
            references = ", ".join(details["refs"])
            return ("A reference can't be configured using name of Mimeo Utils "
                    "or existing Vars. Please rename following refs: "
                    f"[{references}]")

        msg = f"Provided error code is not a {cls.__name__}.Code enum!"
        raise ValueError(msg)


class InvalidMimeoModelError(Exception):
    """A custom Exception class for invalid model configuration.

    Raised when a Mimeo Model is not configured properly.
    """

    class Code(Enum):
        """An Enumeration class for InvalidMimeoModelError error codes.

        Attributes
        ----------
        ERR_1: str
            An error code for missing root in the Mimeo Model
        ERR_2: str
            An error code for multiple roots in the Mimeo Model
        ERR_2: str
            An error code for invalid context name in Mimeo Model
        """

        ERR_1: str = "MISSING_ROOT"
        ERR_2: str = "MULTIPLE_ROOTS"
        ERR_3: str = "INVALID_CONTEXT_NAME"

    def __init__(
            self,
            code: InvalidMimeoModelError.Code,
            **kwargs,
    ):
        """Initialize InvalidMimeoModelError exception with details.

        Extends Exception constructor with a custom message. The message depends on
        an internal InvalidMimeoModelError code.

        Parameters
        ----------
        code : InvalidMimeoModelError.Code
            An internal error code
        kwargs
            An error details
        """
        msg = self._get_msg(code, kwargs)
        super().__init__(msg)

    @classmethod
    def _get_msg(
            cls,
            code: InvalidMimeoModelError.Code,
            details: dict,
    ):
        """Return a custom message based on an error code.

        Parameters
        ----------
        code : InvalidMimeoModelError.Code
            An internal error code
        details : dict
            An error details

        Returns
        -------
        str
            A custom error message

        Raises
        ------
        ValueError
            If the code argument is not InvalidMimeoModelError.Code enum
        """
        if code == cls.Code.ERR_1:
            return f"No root data in Mimeo Model: {details['model']}"
        if code == cls.Code.ERR_2:
            return f"Multiple root data in Mimeo Model: {details['model']}"
        if code == cls.Code.ERR_3:
            return (f"Invalid context name in Mimeo Model (not a string value): "
                    f"{details['model']}")

        msg = f"Provided error code is not a {cls.__name__}.Code enum!"
        raise ValueError(msg)


class InvalidMimeoTemplateError(Exception):
    """A custom Exception class for invalid template configuration.

    Raised when a Mimeo Template is not configured properly.
    """

    def __init__(
            self,
            property_name: str,
            template: dict,
    ):
        """Initialize InvalidMimeoTemplateError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        property_name : str
            A missing property
        template : dict
            A Mimeo Template
        """
        super().__init__(f"No {property_name} property in the Mimeo Template: "
                         f"{template}")


class InvalidMimeoConfigError(Exception):
    """A custom Exception class for invalid mimeo configuration.

    Raised when a Mimeo Configuration is not configured properly.
    """

    class Code(Enum):
        """An Enumeration class for InvalidMimeoConfigError error codes.

        Attributes
        ----------
        ERR_1: str
            An error code for missing templates in the Mimeo Configuration
        ERR_2: str
            An error code for invalid templates in the Mimeo Configuration
        """

        ERR_1: str = "MISSING_TEMPLATES"
        ERR_2: str = "NOT_AN_ARRAY"

    def __init__(
            self,
            code: InvalidMimeoConfigError.Code,
            **kwargs,
    ):
        """Initialize InvalidMimeoConfigError exception with details.

        Extends Exception constructor with a custom message. The message depends on
        an internal InvalidMimeoConfigError code.

        Parameters
        ----------
        code : InvalidMimeoConfigError.Code
            An internal error code
        kwargs
            An error details
        """
        msg = self._get_msg(code, kwargs)
        super().__init__(msg)

    @classmethod
    def _get_msg(
            cls,
            code: InvalidMimeoConfigError.Code,
            details: dict,
    ):
        """Return a custom message based on an error code.

        Parameters
        ----------
        code : InvalidMimeoConfigError.Code
            An internal error code
        details : dict
            An error details

        Returns
        -------
        str
            A custom error message

        Raises
        ------
        ValueError
            If the code argument is not InvalidMimeoConfigError.Code enum
        """
        if code == cls.Code.ERR_1:
            return f"No templates in the Mimeo Config: {details['config']}"
        if code == cls.Code.ERR_2:
            return f"_templates_ property does not store an array: {details['config']}"

        msg = f"Provided error code is not a {cls.__name__}.Code enum!"
        raise ValueError(msg)


class UnsupportedMimeoConfigSourceError(Exception):
    """A custom Exception class for unsupported Mimeo Configuration source.

    Raised when an XML Mimeo Configuration's source is neither a dict nor str.
    """

    def __init__(
            self,
            source: Any,
    ):
        """Initialize UnsupportedMimeoConfigSourceError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        source : Any
            An actual root name
        """
        super().__init__(f"Mimeo Configuration source [{source}] is unsupported. "
                         "Use a file path, stringified configuration or a dictionary.")


class MimeoConfigurationNotFoundError(Exception):
    """A custom Exception class for no mimeo_configuration node in source xml.

    Raised when an XML Mimeo Configuration's root node is not <mimeo_configuration/>.
    """

    def __init__(
            self,
            root_name: str,
    ):
        """Initialize MimeoConfigurationNotFoundError exception with details.

        Extends Exception constructor with a custom message.

        Parameters
        ----------
        root_name : str
            An actual root name
        """
        super().__init__(f"<mimeo_configuration/> not found! {root_name} is not "
                         f"a proper Mimeo Configuration root node.")
